# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda views for management rules"""

import json


from logilab.mtconverter import xml_escape

from cubicweb.predicates import is_instance
from cubicweb_web import INTERNAL_FIELD_VALUE
from cubicweb_web.views import reledit, uicfg

from ..entities import parent_and_container, rule_type_from_etype
from . import viewlib
from . import uicfg as sedauicfg  # noqa - ensure those rules are defined first


affk = uicfg.autoform_field_kwargs


def rule_ref_vocabulary(form, field):
    """Form vocabulary function, for fields referencing a management rules (i.e. a concept in a
    scheme defined on the archive transfer)
    """
    req = form._cw
    parent, container = parent_and_container(form.edited_entity)
    assert container is not None
    parent_etype = form.edited_entity.cw_etype
    if parent_etype == "SEDARefNonRuleId":
        if parent is not None:
            parent_etype = parent.cw_etype
        else:
            parent_etype = json.loads(req.form["arg"][1])
    return _rule_ref_vocabulary(container, parent_etype)


def _rule_ref_vocabulary(container, parent_etype):
    req = container._cw
    rule_type = rule_type_from_etype(parent_etype)
    if container.cw_etype == "SEDAArchiveTransfer":
        rql = (
            "Any C WHERE C in_scheme CS, AT eid %(at)s, "
            "CACLV seda_{0}_rule_code_list_version_from AT, "
            "CACLV seda_{0}_rule_code_list_version_to CS".format(rule_type)
        )
        rset = req.execute(rql, {"at": container.eid})
    else:
        # component archive unit, suppose we can use default vocabularies
        assert rule_type in ("access", "appraisal")
        rql = (
            "Any C WHERE C in_scheme CS, CS scheme_relation_type RT, RT name %(rt)s,"
            "CS scheme_entity_type ET, ET name %(et)s"
        )
        etype = f"SEDASeq{rule_type.capitalize()}RuleRule"
        rset = req.execute(rql, {"rt": "seda_rule", "et": etype})
    if rset:
        return [(req._("<no value specified>"), INTERNAL_FIELD_VALUE)] + sorted(
            (concept.label(), str(concept.eid)) for concept in rset.entities()
        )
    else:
        scheme_relation = f"seda_{rule_type}_rule_code_list_version_from_object"
        scheme_relation = req._(scheme_relation)
        msg = req._("you must specify a scheme for {0} to select a value").format(
            scheme_relation
        )
        return [(msg, INTERNAL_FIELD_VALUE)]


affk.tag_subject_of(
    ("*", "seda_ref_non_rule_id_to", "*"),
    {"choices": rule_ref_vocabulary, "sort": False},
)
affk.tag_subject_of(
    ("*", "seda_rule", "*"), {"choices": rule_ref_vocabulary, "sort": False}
)


class RuleComplexLinkEntityAttributeView(viewlib.TextEntityAttributeView):
    __select__ = viewlib.ComplexLinkEntityAttributeView.__select__ & is_instance(
        "SEDAAccessRule", "SEDADisseminationRule", "SEDAReuseRule", "SEDAHoldRule"
    )

    def entity_call(self, entity):
        _ = self._cw._
        for rule_seq in entity.rules:
            if rule_seq.seda_rule:
                rule = rule_seq.seda_rule[0].label()
                rule = _("rule: {0}").format(rule)
            else:
                rule = _("<no rule specified>")
            if rule_seq.reverse_seda_start_date:
                start_date = _("with {0} start date").format(
                    rule_seq.reverse_seda_start_date[0].user_cardinality
                )
            else:
                start_date = _("without start date")
            self.w(
                "<div>{} {}, {}</div>".format(
                    xml_escape(rule),
                    rule_seq.view("seda.xsdmeta"),
                    xml_escape(start_date),
                )
            )
        if entity.inheritance_control:
            ctrl = entity.inheritance_control
            alternatives = []
            if ctrl.reverse_seda_prevent_inheritance:
                prevent = ctrl.reverse_seda_prevent_inheritance[0]
                # if prevent.prevent_inheritance is None:
                #     alternatives.append(_('XXX prevent to be specified'))
                if prevent.prevent_inheritance:
                    alternatives.append(xml_escape(_("prevent all rules")))
                elif prevent.prevent_inheritance is False:
                    alternatives.append(xml_escape(_("don't prevent inheritance")))
                else:  # unspecified
                    alternatives.append(
                        xml_escape(_("prevent inheritance or not to be specified"))
                    )
            non_rules = []
            for non_rule in ctrl.reverse_seda_ref_non_rule_id_from:
                if non_rule.seda_ref_non_rule_id_to:
                    rule_name = non_rule.seda_ref_non_rule_id_to[0].label()
                    non_rules.append(
                        xml_escape(_("prevent rule {0}").format(rule_name))
                    )
                else:
                    non_rules.append(
                        xml_escape(_("<no rule to prevent specified>"))
                        + " "
                        + non_rule.view("seda.xsdmeta")
                    )
            if non_rules:
                alternatives.append(", ".join(non_rules))
            self.w(f"<b>{self._cw._(' ALT_I18N ')}</b>".join(alternatives))


class FinalActionRuleComplexLinkEntityAttributeView(RuleComplexLinkEntityAttributeView):
    __select__ = viewlib.ComplexLinkEntityAttributeView.__select__ & is_instance(
        "SEDAStorageRule", "SEDAAppraisalRule"
    )

    def entity_call(self, entity):
        _ = self._cw._
        if entity.seda_final_action:
            action = entity.seda_final_action[0].label()
            final_action = _("final action: {0}").format(action)
        else:
            final_action = _("<no final action specified>")
        self.w(f"<div>{xml_escape(final_action)} {entity.view('seda.xsdmeta')}</div>")
        super().entity_call(entity)


class ClassificationRuleComplexLinkEntityAttributeView(
    RuleComplexLinkEntityAttributeView
):
    __select__ = viewlib.ComplexLinkEntityAttributeView.__select__ & is_instance(
        "SEDAClassificationRule"
    )

    def entity_call(self, entity):
        _ = self._cw._
        if entity.seda_classification_level:
            classification_level = entity.seda_classification_level[0].label()
            classification_level = _("classification level: {0}").format(
                classification_level
            )
        else:
            classification_level = _("<no classification level specified>")
        self.w(
            "<div>{} {}</div>".format(
                xml_escape(classification_level), entity.view("seda.xsdmeta")
            )
        )
        if entity.classification_owner:
            classification_owner = entity.classification_owner
            classification_owner = _("classification owner: {0}").format(
                classification_owner
            )
        else:
            classification_owner = _("<no classification owner specified>")
        self.w(f"<div>{xml_escape(classification_owner)}</div>")
        if entity.reverse_seda_classification_reassessing_date:
            reassessing_date = _("with {0} reassessing date").format(
                entity.reverse_seda_classification_reassessing_date[0].user_cardinality
            )
        else:
            reassessing_date = _("without reassessing date")
        self.w(f"<div>{xml_escape(reassessing_date)}</div>")

        if entity.reverse_seda_need_reassessing_authorization:
            reassessing_authorization = (
                entity.reverse_seda_need_reassessing_authorization[0]
            )
            # if reassessing_authorization.need_reassessing_authorization is None:
            #     XXX
            if reassessing_authorization.need_reassessing_authorization:
                need_human = _("need human intervention")
            else:
                need_human = _("without human intervention")
            self.w(f"<div>{xml_escape(need_human)}</div>")
        super().entity_call(entity)


# class HoldRuleComplexLinkEntityAttributeView(RuleComplexLinkEntityAttributeView):
#     __select__ = viewlib.ComplexLinkEntityAttributeView.__select__ & is_instance(
#         "SEDAHoldRule"
#     )

#     def entity_call(self, entity):
#         _ = self._cw._
#         if entity.seda_hold_reason:
#             hold_reason = entity.seda_hold_reason[0].label()
#             hold_reason = _("hold reason: {0}").format(hold_reason)
#         else:
#             hold_reason = _("<no hold reason specified>")
#         self.w(
#             "<div>{} {}</div>".format(
#                 xml_escape(hold_reason), entity.view("seda.xsdmeta")
#             )
#         )
#         if entity.seda_hold_owner:
#             hold_owner = entity.classification_owner
#             hold_owner = _("classification owner: {0}").format(
#                 hold_owner
#             )
#         else:
#             hold_owner = _("<no hold owner specified>")
#         self.w(f"<div>{xml_escape(hold_owner)}</div>")
#         if entity.reverse_seda_classification_reassessing_date:
#             reassessing_date = _("with {0} reassessing date").format(
#                 entity.reverse_seda_classification_reassessing_date[0].user_cardinality
#             )
#         else:
#             reassessing_date = _("without reassessing date")
#         self.w(f"<div>{xml_escape(reassessing_date)}</div>")

#         if entity.reverse_seda_need_reassessing_authorization:
#             reassessing_authorization = (
#                 entity.reverse_seda_need_reassessing_authorization[0]
#             )
#             # if reassessing_authorization.need_reassessing_authorization is None:
#             #     XXX
#             if reassessing_authorization.need_reassessing_authorization:
#                 need_human = _("need human intervention")
#             else:
#                 need_human = _("without human intervention")
#             self.w(f"<div>{xml_escape(need_human)}</div>")
#         super().entity_call(entity)


class MgmtRuleAutoClickAndEditFormView(reledit.AutoClickAndEditFormView):
    __select__ = reledit.AutoClickAndEditFormView.__select__ & is_instance(
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement"
    )

    def _compute_default_value(self, rschema, role):
        if rschema.type.endswith("_rule"):
            rule_type = rschema.type[len("seda_") : -len("_rule")]
            rule = self.entity.inherited_rule(rule_type)
            if rule:
                return rule.view("seda.reledit.text") + "<p><mark>{}</mark></p>".format(
                    self._cw._("inherited value")
                )
        return super()._compute_default_value(rschema, role)



from cubicweb_seda import iter_all_rdefs

for rdef, role in iter_all_rdefs(schema, "SEDAArchiveTransfer"):
    if role == "subject":
        target_etype = rdef.subject
    else:
        target_etype = rdef.object
    if target_etype == "SEDAArchiveUnit":
        sync_schema_props_perms((rdef.subject, rdef.rtype, rdef.object))

sync_schema_props_perms("clone_of")

add_relation_type("code_keyword_type")

scheme = cnx.find("ConceptScheme", title="SEDA 2 : Types de mot-clé").one()
with cnx.deny_all_hooks_but():
    scheme.cw_set(title="SEDA : Types de mot-clé")
    cnx.commit()
    cnx.execute(
        'SET L language_code "seda" WHERE L label_of C, C in_scheme CS, CS eid %(cs)s',
        {"cs": scheme.eid},
    )
    cnx.commit()

for concept in scheme.reverse_in_scheme:
    label = {
        "corpname": "Collectivité",
        "famname": "Nom de famille",
        "geogname": "Nom géographique",
        "name": "Nom",
        "occupation": "Fonction",
        "persname": "Nom de personne",
        "subject": "Mot-matière",
        "genreform": "Typologie documentaire",
        "function": "Activité",
    }[concept.label("seda")]
    cnx.create_entity(
        "Label", label_of=concept, label=label, kind="preferred", language_code="fr"
    )

cnx.commit()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda views for ArchiveUnit"""

from logilab.mtconverter import xml_escape
from logilab.common.registry import objectify_predicate
from logilab.common.decorators import monkeypatch

from cubicweb import tags, _
from cubicweb.predicates import (
    is_instance,
    match_form_params,
    specified_etype_implements,
)
from cubicweb.view import EntityView
from cubicweb.web.views import autoform, baseviews, tabs, uicfg, editforms

from cubicweb_compound import views as compound
from cubicweb_relationwidget import views as rwdg

from ..xsd import un_camel_case
from ..entities import (
    is_full_seda2_profile,
    full_seda2_profile,
    simplified_profile,
    parent_and_container,
)
from ..entities.itree import parent_archive_unit
from . import (
    CONTENT_ETYPE,
    add_subobject_link,
    rtags_from_xsd_element,
    rtags_from_rtype_role_targets,
    has_rel_perm,
)
from . import clone, viewlib, widgets
from . import uicfg as sedauicfg  # noqa - ensure those rules are defined first


def add_links_with_custom_arguments(seq, rtype, role, extraurlparams, sub_unit_types):
    """Return a list of links to be inserted in dropdown button to add subobjects through `rtype`
    and `role` but with an extra argument in url params, given through the `param_defs` list of
    3-uple defining argument's (name, value, link label).
    """
    _ = seq._cw._
    links = []
    urlparams = extraurlparams.copy()
    for argument, value, label in sub_unit_types:
        urlparams[argument] = value
        link = add_subobject_link(seq, rtype, role, urlparams, msg=_(label), klass="")
        if link:
            links.append(link)
    return links


@objectify_predicate
def is_archive_unit_ref(cls, req, rset=None, entity=None, **kwargs):
    """Return 1 if a unit_type value is specified in kwargs or in form parameters, and its value is
    'unit_ref'
    """
    try:
        # first check for unit_type specified in form params
        unit_type = req.form["unit_type"]
        return int(unit_type == "unit_ref")
    except KeyError:
        # if it's not, look if we're in the context of a archive unit or its first level choice
        if entity is None:
            assert rset is not None, (
                "is_archive_unit_ref can only be used in the context of a SEDAArchiveUnit "
                "or SEDAAltArchiveUnitArchiveUnitRefId entity, but no context entity specified"
            )
            entity = rset.get_entity(0, 0)
        if entity.cw_etype == "SEDAArchiveUnit":
            entity = entity.first_level_choice
        elif entity.cw_etype == "SEDAArchiveUnitRefId":
            if not entity.seda_archive_unit_ref_id_from:
                return 0  # other kind of reference
            entity = entity.seda_archive_unit_ref_id_from[0]
            if entity.cw_etype != "SEDAAltArchiveUnitArchiveUnitRefId":
                return 0  # other kind of reference
        assert entity.cw_etype == "SEDAAltArchiveUnitArchiveUnitRefId", (
            "is_archive_unit_ref can only be used in the context of a SEDAArchiveUnit, "
            "SEDAArchiveUnitRefId or SEDAAltArchiveUnitArchiveUnitRefId entity, not %s"
            % entity
        )
        return 0 if entity.content_sequence else 1


@objectify_predicate
def is_simple_keyword(cls, req, rset=None, entity=None, **kwargs):
    """Return 1 if a keyword_type value is specified in kwargs or in form
    parameters, and its value is 'simple_keyword', or if given entity is not
    linked to a keyword reference."""
    try:
        # first check for unit_type specified in form params
        unit_type = req.form["keyword_type"]
        return int(unit_type == "simple_keyword")
    except KeyError:
        if entity is not None and not entity.reverse_seda_keyword_reference_from:
            return 1
        return 0


@objectify_predicate
def typed_reference(cls, req, entity=None, **kwargs):
    """Return positive score for content's typed data object references (IsPartOf, VersionOf, etc.),
    not those starting directly from archive unit.
    """
    # code below don't handle properly the case where entity isn't yet created while rtype isn't
    # specified (e.g. because it's an inlined form). To avoid false positive, we've to also ensure
    # we're within a full seda 2 profile (typed references aren't allowed in simplified profiles)
    if not is_full_seda2_profile(entity, kwargs.get("rset")):
        return 0
    if entity is None or not entity.has_eid():
        try:
            rtype, eid, role = req.form["__linkto"].split(":")
        except KeyError:
            pass
        else:
            if rtype == "seda_data_object_reference":
                entity = req.entity_from_eid(eid)
            else:
                return 0
    else:
        entity = entity.seda_data_object_reference[0]
    if entity is not None and entity.cw_etype == CONTENT_ETYPE:
        return 0
    return 1


afs = uicfg.autoform_section
affk = uicfg.autoform_field_kwargs
pvs = uicfg.primaryview_section
pvdc = uicfg.primaryview_display_ctrl
rec = uicfg.reledit_ctrl


def unit_ref_vocabulary(form, field):
    """Form vocabulary function for archive unit references, necessary to get parent container while
    the entity is being created.
    """
    req = form._cw
    parent, container = parent_and_container(form.edited_entity)
    assert container is not None
    rset = req.execute(
        "Any X, XUA ORDERBY XUA WHERE "
        "X is SEDAArchiveUnit, X user_annotation XUA, X container R, R eid %(root)s",
        {"root": container.eid},
    )
    return [(entity.dc_title(), str(entity.eid)) for entity in rset.entities()]


affk.tag_subject_of(
    ("SEDAArchiveUnitRefId", "seda_archive_unit_ref_id_to", "*"),
    {"choices": unit_ref_vocabulary},
)


def do_ref_vocabulary(form, field):
    """Form vocabulary function for data object references, necessary to get parent container while
    the entity is being created.
    """
    req = form._cw
    parent, container = parent_and_container(form.edited_entity)
    assert container is not None
    rset = req.execute(
        "Any X, XUA ORDERBY XUA WHERE "
        "X is IN(SEDABinaryDataObject, SEDAPhysicalDataObject), "
        "X user_annotation XUA, X container R, R eid %(root)s",
        {"root": container.eid},
    )
    return [(entity.dc_title(), str(entity.eid)) for entity in rset.entities()]


for rtype in ("seda_data_object_reference_id", "seda_target"):
    affk.tag_subject_of(("*", rtype, "*"), {"choices": do_ref_vocabulary})


pvs.tag_subject_of(
    ("SEDAArchiveUnit", "seda_alt_archive_unit_archive_unit_ref_id", "*"), "hidden"
)


class SkipIBreadCrumbsAdapter(compound.IContainedBreadcrumbsAdapter):
    """IBreadCrumbsAdapter for entities which should not appears in breadcrumb,
    we want to go back to the parent
    """

    __select__ = is_instance("SEDAAltArchiveUnitArchiveUnitRefId", CONTENT_ETYPE)

    def breadcrumbs(self, view=None, recurs=None):
        parent = self.parent_entity()
        adapter = parent.cw_adapt_to("IBreadCrumbs")
        return adapter.breadcrumbs(view, recurs)


class SkipInContextView(baseviews.InContextView):
    """Custom incontext view, for use in title of creation form, among others"""

    __select__ = is_instance("SEDAAltArchiveUnitArchiveUnitRefId", CONTENT_ETYPE)

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        parent_archive_unit(entity).view("incontext", w=self.w)


class ArchiveUnitSubObjectsTab(viewlib.SubObjectsTab):
    """Abstract subobjects tab specific to archive unit to handle subentities below choice>sequence
    child transparently.
    """

    __abstract__ = True
    __select__ = is_instance("SEDAArchiveUnit") & ~is_archive_unit_ref()

    def entity_call(self, entity):
        seq = entity.first_level_choice.content_sequence
        self.display_add_button(seq)
        self.display_subobjects_list(seq)

    def url_params(self, entity):
        archive_unit = parent_archive_unit(entity)
        return {
            "__redirectparams": "tab=" + self.__regid__,
            "__redirectpath": archive_unit.rest_path(),
        }

    def parent(self, entity):
        return parent_archive_unit(entity)


class ArchiveUnitContentTab(tabs.TabsMixin, EntityView):
    """Base class for tabs whose content is generated by redirecting to a view on archive unit's
    content.
    """

    __abstract__ = True
    __select__ = ArchiveUnitSubObjectsTab.__select__
    content_vid = None  # set in concrete class

    def entity_call(self, entity):
        seq = entity.first_level_choice.content_sequence
        seq.view(self.content_vid, w=self.w)


class ContentSubObjectsView(viewlib.SubObjectsTab):
    """Base class to display subobjects of archive unit's content as if they were directly linked
    to it.
    """

    __abstract__ = True
    __select__ = is_instance(CONTENT_ETYPE)
    tabid = None  # set in concrete class

    def url_params(self, entity):
        params = super().url_params(entity)
        params["__redirectpath"] = self.parent(entity).rest_path()
        return params

    def parent(self, entity):
        return entity.reverse_seda_seq_alt_archive_unit_archive_unit_ref_id_management[
            0
        ].reverse_seda_alt_archive_unit_archive_unit_ref_id[0]


class ArchiveUnitTabbedPrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance("SEDAArchiveUnit")

    tabs = [
        "main_tab",
        _("seda_management_tab"),
        _("seda_archive_units_tab"),
        _("seda_data_objects_tab"),
        # Content tabs
        _("seda_identification_tab"),
        _("seda_dates_tab"),
        _("seda_gps_tab"),
        _("seda_services_tab"),
        _("seda_agents_tab"),
        _("seda_coverage_tab"),
        _("seda_indexation_tab"),
        _("seda_relations_tab"),
        _("seda_events_tab"),
        _("seda_history_tab"),
        _("seda_email_tab"),
    ]

    def entity_call(self, entity, **kwargs):
        super().entity_call(entity, **kwargs)
        rwdg.bootstrap_dialog(self.w, self._cw._, clone._import_div_id(entity))


# main tab for archive unit reference ##########################################

au_ref_pvs = pvs.derive(
    __name__, is_instance("SEDAArchiveUnit") & is_archive_unit_ref()
)
au_ref_pvs.tag_subject_of(
    ("SEDAArchiveUnit", "seda_alt_archive_unit_archive_unit_ref_id", "*"), "attributes"
)
rec.tag_subject_of(
    ("SEDAArchiveUnit", "seda_alt_archive_unit_archive_unit_ref_id", "*"),
    {"rvid": "seda.reledit.simplelink", "novalue_label": " "},
)


class RefArchiveUnitSimpleLinkToEntityAttributeView(EntityView):
    __regid__ = "seda.reledit.simplelink"
    __select__ = (
        is_instance("SEDAAltArchiveUnitArchiveUnitRefId") & is_archive_unit_ref()
    )

    def entity_call(self, entity):
        entity.reference.view(
            "seda.reledit.complexlink",
            initargs={"rtype": "seda_archive_unit_ref_id_from"},
            w=self.w,
        )


# management tab ###############################################################


class ArchiveUnitManagementTab(viewlib.PrimaryTabWithoutBoxes):
    """Display management information about an archive unit."""

    __regid__ = "seda_management_tab"
    __select__ = is_instance("SEDAArchiveUnit") & ~is_archive_unit_ref()

    rtype_role_targets = [
        ("seda_storage_rule", "object", None),
        ("seda_appraisal_rule", "object", None),
        ("seda_access_rule", "object", None),
        ("seda_dissemination_rule", "object", None),
        ("seda_reuse_rule", "object", None),
        ("seda_hold_rule", "object", None),
        ("seda_classification_rule", "object", None),
        ("seda_need_authorization", "object", None),
    ]
    rsection, display_ctrl = rtags_from_rtype_role_targets(
        CONTENT_ETYPE, rtype_role_targets
    )

    def entity_call(self, entity, **kwargs):
        seq = entity.first_level_choice.content_sequence
        super().entity_call(seq, **kwargs)


class SimplifiedArchiveUnitManagementTab(ArchiveUnitManagementTab):
    __select__ = ArchiveUnitManagementTab.__select__ & simplified_profile()

    rtype_role_targets = [
        ("seda_appraisal_rule", "object", None),
        ("seda_access_rule", "object", None),
    ]
    rsection, display_ctrl = rtags_from_rtype_role_targets(
        CONTENT_ETYPE, rtype_role_targets
    )


# identification tab ###########################################################


class ArchiveUnitIdentificationTab(ArchiveUnitContentTab):
    """Display identification information about a full seda2 archive unit."""

    __regid__ = "seda_identification_tab"
    __select__ = ArchiveUnitContentTab.__select__ & full_seda2_profile()
    content_vid = "seda_content_identification"


class ContentIdentificationView(viewlib.PrimaryTabWithoutBoxes):
    """Display identification information about an archive unit content."""

    __regid__ = "seda_content_identification"
    __select__ = is_instance(CONTENT_ETYPE)

    rtype_role_targets = [
        ("seda_source", "object", None),
        ("seda_file_plan_position", "object", None),
        ("seda_system_id", "object", None),
        ("seda_originating_system_id", "object", None),
        ("seda_archival_agency_archive_unit_identifier", "object", None),
        ("seda_originating_agency_archive_unit_identifier", "object", None),
        ("seda_transferring_agency_archive_unit_identifier", "object", None),
    ]
    rsection, display_ctrl = rtags_from_rtype_role_targets(
        CONTENT_ETYPE, rtype_role_targets
    )


# dates tab ####################################################################


class ArchiveUnitDatesTab(ArchiveUnitContentTab):
    """Display dates information about a full seda2 archive unit."""

    __regid__ = "seda_dates_tab"
    __select__ = ArchiveUnitContentTab.__select__ & full_seda2_profile()
    content_vid = "seda_content_dates"


class ContentDatesView(viewlib.PrimaryTabWithoutBoxes):
    """Display dates information about an archive unit content."""

    __regid__ = "seda_content_dates"
    __select__ = is_instance(CONTENT_ETYPE)

    rtype_role_targets = [
        ("seda_created_date", "object", None),
        ("seda_transacted_date", "object", None),
        ("seda_acquired_date", "object", None),
        ("seda_sent_date", "object", None),
        ("seda_received_date", "object", None),
        ("seda_registered_date", "object", None),
        ("seda_date_litteral", "object", None),
    ]
    rsection, display_ctrl = rtags_from_rtype_role_targets(
        CONTENT_ETYPE, rtype_role_targets
    )


# courriel/email tab ###########################################################


class ArchiveUnitEmailTab(ArchiveUnitContentTab):
    """Display dates information about a full seda2 archive unit."""

    __regid__ = "seda_email_tab"
    __select__ = ArchiveUnitContentTab.__select__ & full_seda2_profile()
    content_vid = "seda_content_email"


class ContentEmailView(viewlib.PrimaryTabWithoutBoxes):
    """Display email information about an archive unit content."""

    __regid__ = "seda_content_email"
    __select__ = is_instance(CONTENT_ETYPE)

    rtype_role_targets = [
        ("seda_originating_system_id_reply_to", "object", None),
        ("seda_text_content", "object", None),
    ]
    rsection, display_ctrl = rtags_from_rtype_role_targets(
        CONTENT_ETYPE, rtype_role_targets
    )


# GPS tab ######################################################################


class ArchiveUnitGPSTab(ArchiveUnitContentTab):
    """Display GPS information about a full seda2 archive unit."""

    __regid__ = "seda_gps_tab"
    __select__ = ArchiveUnitContentTab.__select__ & full_seda2_profile()
    content_vid = "seda_content_gps"


class ContentGPSView(viewlib.PrimaryTabWithoutBoxes):
    """Display GPS information about an archive unit content."""

    __regid__ = "seda_content_gps"
    __select__ = is_instance(CONTENT_ETYPE)

    rsection, display_ctrl = rtags_from_xsd_element(CONTENT_ETYPE, "Gps")


# services tab #################################################################


class ArchiveUnitServicesTab(ArchiveUnitContentTab):
    """Display services related to a full seda2 archive unit."""

    __regid__ = "seda_services_tab"
    __select__ = ArchiveUnitContentTab.__select__ & full_seda2_profile()
    content_vid = "seda_content_services"


class ContentServicesView(viewlib.PrimaryTabWithoutBoxes):
    """Display services related to an archive unit content."""

    __regid__ = "seda_content_services"
    __select__ = is_instance(CONTENT_ETYPE)

    rtype_role_targets = [
        ("seda_originating_agency_from", "object", None),
        ("seda_submission_agency_from", "object", None),
        ("seda_authorized_agent_from", "object", None),
    ]
    rsection, display_ctrl = rtags_from_rtype_role_targets(
        CONTENT_ETYPE, rtype_role_targets
    )


# agents tab ###################################################################


class ArchiveUnitAgentsTab(ArchiveUnitContentTab):
    """Display agents related to a full seda2 archive unit."""

    __regid__ = "seda_agents_tab"
    __select__ = ArchiveUnitContentTab.__select__ & full_seda2_profile()
    content_vid = "seda_content_agents"


class ContentAgentsView(ContentSubObjectsView):
    """Display agents related to an archive unit content."""

    __regid__ = "seda_content_agents"
    __select__ = is_instance(CONTENT_ETYPE)
    tabid = ArchiveUnitAgentsTab.__regid__

    rtype_role_targets = [
        ("seda_writer_from", "object", None),
        ("seda_addressee_from", "object", None),
        ("seda_recipient_from", "object", None),
        ("seda_transmitter", "object", None),
        ("seda_sender", "object", None),
        ("seda_agent", "object", None),
    ]

    _(
        "creating SEDAWriter (SEDAWriter seda_writer_from "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDAAddressee (SEDAAddressee seda_addressee_from "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDARecipient (SEDARecipient seda_recipient_from "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDATransmitter (SEDATransmitter seda_transmitter "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDASender (SEDASender seda_sender "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDAAgent (SEDAAgent seda_agent "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )


# coverage tab #################################################################


class ArchiveUnitCoverageTab(ArchiveUnitContentTab):
    """Display coverage information about a full seda2 archive unit."""

    __regid__ = "seda_coverage_tab"
    __select__ = ArchiveUnitContentTab.__select__ & full_seda2_profile()
    content_vid = "seda_content_coverage"


class ContentCoverageView(ContentSubObjectsView):
    """Display coverage information about an archive unit content."""

    __regid__ = "seda_content_coverage"
    __select__ = is_instance(CONTENT_ETYPE)
    tabid = ArchiveUnitCoverageTab.__regid__

    rtype_role_targets = [
        ("seda_spatial", "object", None),
        ("seda_temporal", "object", None),
        ("seda_juridictional", "object", None),
    ]

    _(
        "creating SEDASpatial (SEDASpatial seda_spatial "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDATemporal (SEDATemporal seda_temporal "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDAJuridictional (SEDAJuridictional seda_juridictional "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )


class TypeListItemContentView(viewlib.ListItemContentView):
    """Extended list item content view including the entity's type (for case where the list may
    include entities of different types).
    """

    __select__ = is_instance(
        "SEDAWriter",
        "SEDAAddressee",
        "SEDARecipient",
        "SEDASpatial",
        "SEDATemporal",
        "SEDAJuridictional",
        "SEDAAgent",
    )

    def entity_call(self, entity):
        entity.view("seda.type_meta", w=self.w)


# indexation tab ###############################################################


class ArchiveUnitIndexationTab(ArchiveUnitContentTab):
    """Display content's indexation about an archive unit."""

    __regid__ = "seda_indexation_tab"
    content_vid = "seda_content_indexation"


class ContentIndexationView(ContentSubObjectsView):
    """Display indexation (keywords and tags) about an archive unit content."""

    __regid__ = "seda_content_indexation"
    tabid = ArchiveUnitIndexationTab.__regid__

    rtype_role_targets = [
        ("seda_keyword", "object", "SEDAKeyword"),
        ("seda_tag", "object", "SEDATag"),
    ]
    keyword_custom_arguments = [
        ("keyword_type", "concept_keyword", _("keyword concept")),
        ("keyword_type", "simple_keyword", _("keyword simple")),
    ]

    _(
        "creating SEDAKeyword (SEDAKeyword seda_keyword "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDATag (SEDATag seda_tag "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )

    def display_add_button(self, entity):
        urlparams = self.url_params(entity)
        links = add_links_with_custom_arguments(
            entity, "seda_keyword", "object", urlparams, self.keyword_custom_arguments
        )
        rtype_roles = [
            (rtype, role)
            for rtype, role, _ in self.rtype_role_targets
            if rtype != "seda_keyword"
        ]
        links += viewlib.add_links_from_rtype_roles(entity, rtype_roles, urlparams)
        viewlib.display_add_button(self.w, links, entity._cw.__("add"))


class SimplifiedArchiveUnitContentIndexationView(ContentIndexationView):
    __select__ = ContentIndexationView.__select__ & simplified_profile()

    rtype_role_targets = [("seda_keyword", "object", "SEDAKeyword")]


class KeywordBusinessValueEntityView(viewlib.ListItemContentView):
    __select__ = viewlib.ListItemContentView.__select__ & is_instance("SEDAKeyword")

    def entity_call(self, entity):
        if entity.reverse_seda_keyword_reference_from:
            kwr = entity.reverse_seda_keyword_reference_from[0]
            if kwr.concept:
                kwr_value = kwr.concept.view("oneline")
                msg = xml_escape(self._cw._("keyword concept: {0}")).format(kwr_value)
            elif kwr.scheme:
                kwr_value = kwr.scheme.view("oneline")
                msg = xml_escape(self._cw._("keyword scheme: {0}")).format(kwr_value)
            else:
                msg = xml_escape(self._cw._("<no keyword scheme specified>"))
        else:
            if entity.keyword_content:
                content = entity.keyword_content
            else:
                content = self._cw._("<no value specified>")
            msg = xml_escape(self._cw._("keyword: {0}").format(content))
        self.w(f"<span class=\"value\">{msg} {entity.view('seda.xsdmeta')}</span>")

        if entity.reverse_seda_keyword_type_from:
            kwt = entity.reverse_seda_keyword_type_from[0]
            if kwt.seda_keyword_type_to:
                kwt_value = kwt.seda_keyword_type_to[0].label()
            else:
                kwt_value = self._cw._("<no type specified>")
            msg = xml_escape(self._cw._("keyword type: {0}").format(kwt_value))
            self.w(f"<br/><span>{msg} {kwt.view('seda.xsdmeta')}</span>")


afs.tag_subject_of(
    ("SEDAKeywordReference", "seda_keyword_reference_to_scheme", "*"),
    "main",
    "attributes",
)
affk.set_field_kwargs(
    "SEDAKeywordReference",
    "seda_keyword_reference_to_scheme",
    help=_("select a type to filter out unrelated vocabularies"),
)
affk.set_field_kwargs(
    "SEDAKeywordReference",
    "seda_keyword_reference_to",
    help=_("select first a vocabulary then type here to prompt auto-completion"),
    widget=widgets.ConceptAutoCompleteWidget(
        slave_name="seda_keyword_reference_to",
        master_name="seda_keyword_reference_to_scheme",
    ),
)
affk.set_fields_order(
    "SEDAKeywordReference",
    [
        "user_cardinality",
        "seda_keyword_reference_to_scheme",
        "seda_keyword_reference_to",
    ],
)

# add rule in afs for 'complex keyword'
afs.tag_attribute(("SEDAKeyword", "keyword_content"), "main", "hidden")
afs.tag_attribute(("SEDAKeywordReference", "user_cardinality"), "main", "hidden")
# custom afs to handle 'simple keyword'
kw_simple_afs = afs.derive(__name__, is_instance("SEDAKeyword") & is_simple_keyword())
kw_simple_afs.tag_attribute(("SEDAKeyword", "keyword_content"), "main", "attributes")
kw_simple_afs.tag_object_of(
    ("*", "seda_keyword_reference_from", "SEDAKeyword"), "main", "hidden"
)
# but one ordering is enough to rule them all
affk.set_fields_order(
    "SEDAKeyword",
    [
        ("seda_keyword_type_from", "object"),
        "keyword_content",
        ("seda_keyword_reference_from", "object"),
    ],
)

affk.set_field_kwargs(
    "SEDAKeywordType",
    "seda_keyword_type_to",
    widget=widgets.KeywordTypeMasterWidget(
        slave_base_name="seda_keyword_reference_to_scheme"
    ),
)


class SEDAWarningExportFormViewMixin:
    """Add a warning abouta field beeing not available in seda 2.1"""

    def form_warning(self, entity):
        msg = self._cw._(
            "{0} is not available in seda 2.1 and will not be exported in these profiles."
        ).format(self._cw._(entity.__regid__))
        self.w(f'<div class="alert alert-warning alert-form">{msg}</div>')

    def init_form(self, form, entity):
        """customize your form before rendering here"""
        self.form_warning(entity)
        super().init_form(form, entity)


class SEDAAgentEditionFormView(
    SEDAWarningExportFormViewMixin, editforms.EditionFormView
):
    """Add a warning about SEDAAgent being not available in seda 2.1"""

    __select__ = is_instance("SEDAAgent")


class SEDAAgentCreationFormView(
    SEDAWarningExportFormViewMixin, editforms.CreationFormView
):
    """Add a warning about SEDAAgent beeing not available in seda 2.1"""

    __select__ = specified_etype_implements("SEDAAgent")


class ComplexKeywordAutomaticEntityForm(autoform.AutomaticEntityForm):
    __select__ = is_instance("SEDAKeyword") & ~is_simple_keyword()

    def should_display_inline_creation_form(self, rschema, existing, card):
        # force display of type and keyword
        if not existing and rschema in (
            "seda_keyword_type_from",
            "seda_keyword_reference_from",
        ):
            return True
        return False

    def should_display_add_new_relation_link(self, rschema, existing, card):
        return False


class ComplexKeywordEditionFormView(autoform.InlineEntityEditionFormView):
    __select__ = (
        autoform.InlineEntityEditionFormView.__select__
        & is_instance("SEDAKeywordType", "SEDAKeywordReference")
        & ~is_simple_keyword()
    )
    removejs = None


class ComplexKeywordCreationFormView(autoform.InlineEntityCreationFormView):
    __select__ = (
        autoform.InlineEntityCreationFormView.__select__
        & specified_etype_implements("SEDAKeywordType", "SEDAKeywordReference")
        & ~is_simple_keyword()
    )
    removejs = None


# relations tab ################################################################


class ArchiveUnitRelationsTab(ArchiveUnitContentTab):
    """Display content's relations about a full seda2  archive unit."""

    __regid__ = "seda_relations_tab"
    __select__ = ArchiveUnitContentTab.__select__ & full_seda2_profile()
    content_vid = "seda_content_relations"


class ContentRelationsView(ContentSubObjectsView):
    """Display relation information about an archive unit content."""

    __regid__ = "seda_content_relations"
    tabid = ArchiveUnitRelationsTab.__regid__

    rtype_role_targets = [
        ("seda_is_version_of", "object", None),
        ("seda_replaces", "object", None),
        ("seda_requires", "object", None),
        ("seda_is_part_of", "object", None),
        ("seda_references", "object", None),
    ]

    _(
        "creating SEDAIsVersionOf (SEDAIsVersionOf seda_is_version_of "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDAReplaces (SEDAReplaces seda_replaces "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDARequires (SEDARequires seda_requires "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDAIsPartOf (SEDAIsPartOf seda_is_part_of "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )
    _(
        "creating SEDAReferences (SEDAReferences seda_references "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )


class ReferenceListItemContentView(viewlib.ListItemContentView):
    __select__ = is_instance(
        "SEDAIsVersionOf",
        "SEDAReplaces",
        "SEDARequires",
        "SEDAIsPartOf",
        "SEDAReferences",
    )

    def entity_call(self, entity):
        self.w(f"<b>{entity.dc_type()} ")
        entity.view("seda.xsdmeta")
        self.w("</b> :")
        alt_rtype = f"seda_alt_{un_camel_case(entity.cw_etype[4:])}_archive_unit_ref_id"
        alt = entity.related(alt_rtype).one()
        alternatives = viewlib.alternative_values(alt, alt_rtype)
        self.w('<div class="value alternative">')
        if alternatives:
            self.w(alternatives)
        else:
            self.wdata(self._cw._("<no value specified>"))
        self.w("</div>")
        self.w('<div class="clearfix"/>')


do_ref_afs = afs.derive(
    __name__, is_instance("SEDADataObjectReference") & typed_reference()
)
do_ref_afs.tag_attribute(
    ("SEDADataObjectReference", "user_cardinality"), "main", "hidden"
)

for etype in (
    "SEDAAltIsPartOfArchiveUnitRefId",
    "SEDAAltIsVersionOfArchiveUnitRefId",
    "SEDAAltReferencesArchiveUnitRefId",
    "SEDAAltReplacesArchiveUnitRefId",
    "SEDAAltRequiresArchiveUnitRefId",
):
    affk.set_fields_order(
        etype,
        [
            ("seda_data_object_reference", "object"),
            ("seda_repository_object_pid", "object"),
            ("seda_archive_unit_ref_id_from", "object"),
            ("seda_repository_archive_unit_pid", "object"),
        ],
    )


# event tab ####################################################################


class ArchiveUnitEventsTab(ArchiveUnitContentTab):
    """Display content's relations about a full seda2 archive unit."""

    __regid__ = "seda_events_tab"
    __select__ = ArchiveUnitContentTab.__select__ & full_seda2_profile()
    content_vid = "seda_content_events"


class ContentEventsView(ContentSubObjectsView):
    """Display events about an archive unit content."""

    __regid__ = "seda_content_events"
    tabid = ArchiveUnitEventsTab.__regid__

    rtype_role_targets = [
        ("seda_event", "object", None),
    ]

    _(
        "creating SEDAEvent (SEDAEvent seda_event "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )


class EventListItemContentView(viewlib.ListItemContentView):
    __select__ = viewlib.ListItemContentView.__select__ & is_instance("SEDAEvent")

    def entity_call(self, entity):
        entity.view("seda.xsdmeta", w=self.w, skip_one_card=True, with_annotation=False)
        attrs = []
        for rtype in [
            "seda_event_type_from",
            "seda_event_identifier",
            "seda_event_detail",
        ]:
            related = getattr(entity, "reverse_" + rtype)
            if related:
                if related[0].user_cardinality == "1":
                    card = self._cw._("mandatory")
                else:
                    card = self._cw._("optional")
                value = ""
                if rtype == "seda_event_type_from":
                    value = related[0].seda_event_type_to
                    if value:
                        value = value[0].label()
                attrs.append(f"{self._cw.__(rtype + '_object')} {value} {card}")
        if attrs:
            self.w(f" ({', '.join(attrs)})")
        description = getattr(entity, "user_annotation", None)
        if description:
            self.w(f' <div class="description text-muted">{description}</div>')


affk.set_fields_order(
    "SEDAEvent",
    [
        "user_cardinality",
        ("seda_event_type_from", "object"),
        ("seda_event_identifier", "object"),
        ("seda_event_detail", "object"),
    ],
)


# history tab ##################################################################


class ArchiveUnitHistoryTab(ArchiveUnitContentTab):
    """Display content's custodial history about an archive unit."""

    __regid__ = "seda_history_tab"
    content_vid = "seda_content_history"


class ContentHistoryView(ContentSubObjectsView):
    """Display custodial history information about an archive unit content."""

    __regid__ = "seda_content_history"
    tabid = ArchiveUnitHistoryTab.__regid__

    rtype_role_targets = [("seda_custodial_history_item", "object", None)]

    _(
        "creating SEDACustodialHistoryItem (SEDACustodialHistoryItem seda_custodial_history_item "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )


class CustodialHistoryItemListItemContentView(viewlib.ListItemContentView):
    __select__ = viewlib.ListItemContentView.__select__ & is_instance(
        "SEDACustodialHistoryItem"
    )

    def entity_call(self, entity):
        super().entity_call(entity)
        if entity.reverse_seda_when:
            when = entity.reverse_seda_when[0]
            if when.user_cardinality == "1":
                self.w(f" ({self._cw._('mandatory timestamp')})")
            else:
                self.w(f" ({self._cw._('optional timestamp')})")


# main tab for archive unit content ############################################


class ArchiveUnitPrimaryTab(tabs.PrimaryTab):
    """Overriden primary tab to display attributes from both the archive unit and it's content
    sequence.
    """

    __select__ = is_instance("SEDAArchiveUnit") & ~is_archive_unit_ref()

    def render_entity_attributes(self, entity):
        """Renders all attributes and relations in the 'attributes' section."""
        display_attributes = list(self._display_attributes(entity))
        seq = entity.first_level_choice.content_sequence
        view = self._cw.vreg["views"].select(
            "seda_content_main", self._cw, rset=seq.as_rset()
        )
        display_attributes += view._display_attributes(seq)
        if display_attributes:
            self.w('<table class="table cw-table-primary-entity">')
            for rschema, role, dispctrl, value in display_attributes:
                label = self._rel_label(entity, rschema, role, dispctrl)
                self.render_attribute(label, value, table=True)
            self.w("</table>")


# add as a monkey patch the first step of original 'render_entity_attributes'
# method, which is unfortunatly monolitic in cubicweb. We need this so we can
# have attributes from two distinct entities as if it was only one.


@monkeypatch(tabs.PrimaryTab)
def _display_attributes(self, entity):
    for rschema, targets, role, dispctrl in self._section_def(entity, "attributes"):
        vid = dispctrl.get("vid", "reledit")
        if rschema.final or vid == "reledit" or dispctrl.get("rtypevid"):
            value = entity.view(
                vid, rtype=rschema.type, role=role, initargs={"dispctrl": dispctrl}
            )
        else:
            rset = self._relation_rset(entity, rschema, role, dispctrl)
            if rset:
                value = self._cw.view(vid, rset)
            else:
                value = None
        if value is not None and value != "":
            yield (rschema, role, dispctrl, value)


content_ordered_fields = [
    ("seda_description_level_from", "object"),
    ("seda_title", "object"),
    ("seda_start_date", "object"),
    ("seda_end_date", "object"),
    ("seda_description", "object"),
    ("seda_originating_agency_from", "object"),
    ("seda_transferring_agency_archive_unit_identifier", "object"),
    ("seda_system_id", "object"),
    ("seda_version", "object"),
    ("seda_type_from", "object"),
    ("seda_document_type", "object"),
    ("seda_status", "object"),
    ("seda_language_from", "object"),
    ("seda_description_language_from", "object"),
]


class ContentMainView(viewlib.PrimaryTabWithoutBoxes):
    """Display content information which are not in a dedicated tab."""

    __regid__ = "seda_content_main"
    __select__ = is_instance(CONTENT_ETYPE)

    rtype_role_targets = [field_def + (None,) for field_def in content_ordered_fields]
    rsection, display_ctrl = rtags_from_rtype_role_targets(
        CONTENT_ETYPE, rtype_role_targets
    )


class SimplifiedContentMainView(ContentMainView):
    """Display content information which are not in a dedicated tab for a
    simplified archive unit."""

    __select__ = ContentMainView.__select__ & simplified_profile()

    rtype_role_targets = [
        ("seda_description_level_from", "object", None),
        ("seda_title", "object", None),
        ("seda_start_date", "object", None),
        ("seda_end_date", "object", None),
        ("seda_description", "object", None),
        ("seda_originating_agency_from", "object", None),
        ("seda_transferring_agency_archive_unit_identifier", "object", None),
        ("seda_system_id", "object", None),
        ("seda_language_from", "object", None),
    ]
    rsection, display_ctrl = rtags_from_rtype_role_targets(
        CONTENT_ETYPE, rtype_role_targets
    )


for rtype, role in content_ordered_fields:
    if role == "subject":
        pvs.tag_subject_of((CONTENT_ETYPE, rtype, "*"), "attributes")
    else:
        pvs.tag_object_of(("*", rtype, CONTENT_ETYPE), "attributes")
    if rtype == "seda_description_level_to":
        novalue_label = _("<no value specified>")
    else:
        novalue_label = " "
    vid = "seda.reledit.complexlink" if rtype.endswith("_from") else "seda.reledit.text"
    if role == "subject":
        rec.tag_subject_of(
            (CONTENT_ETYPE, rtype, "*"), {"rvid": vid, "novalue_label": novalue_label}
        )
    else:
        rec.tag_object_of(
            ("*", rtype, CONTENT_ETYPE), {"rvid": vid, "novalue_label": novalue_label}
        )

pvs.tag_object_of(
    ("*", "seda_seq_alt_archive_unit_archive_unit_ref_id_management", "*"), "hidden"
)

for rtype in (
    "seda_description_level_to",
    "seda_language_to",
    "seda_description_language_to",
    "seda_archival_agency",
    "seda_transferring_agency",
    "seda_originating_agency_to",
    "seda_submission_agency_to",
    "seda_authorized_agent_to",
    "seda_writer_to",
    "seda_addressee_to",
    "seda_recipient_to",
):
    affk.tag_subject_of(("*", rtype, "*"), {"widget": rwdg.RelationFacetWidget})

affk.set_fields_order(
    CONTENT_ETYPE, ["user_cardinality", "user_annotation"] + content_ordered_fields
)
pvdc.set_fields_order(
    CONTENT_ETYPE, ["user_cardinality", "user_annotation"] + content_ordered_fields
)

# remove from relations section and autoform what is shown in tabs
for cls in (
    ContentIdentificationView,
    ContentDatesView,
    ContentServicesView,
    ContentAgentsView,
    ContentIndexationView,
    ContentEventsView,
    ContentEmailView,
):
    for rtype, role, target in cls.rtype_role_targets:
        if role == "object":
            pvs.tag_object_of(("*", rtype, CONTENT_ETYPE), "hidden")
            afs.tag_object_of(("*", rtype, CONTENT_ETYPE), "main", "hidden")
        else:
            pvs.tag_subject_of((CONTENT_ETYPE, rtype, "*"), "hidden")
            afs.tag_subject_of((CONTENT_ETYPE, rtype, "*"), "main", "hidden")


# archive units tab ############################################################


class ArchiveUnitArchiveUnitsTab(ArchiveUnitSubObjectsTab):
    """Tab for sub-archive units of an archive unit"""

    __regid__ = "seda_archive_units_tab"

    rtype = "seda_archive_unit"
    role = "object"
    unit_custom_arguments = [
        ("unit_type", "unit_content", _("archive unit (content)")),
        ("unit_type", "unit_ref", _("archive unit (reference)")),
    ]

    _(
        "creating SEDAArchiveUnit (SEDAArchiveUnit seda_archive_unit "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )

    def display_add_button(self, entity):
        urlparams = self.url_params(entity)
        links = add_links_with_custom_arguments(
            entity, self.rtype, self.role, urlparams, self.unit_custom_arguments
        )
        if links:
            import_url = clone.import_unit_url(self._cw, parent_archive_unit(entity))
            links.append(tags.a(self._cw._("import_unit"), href=import_url))
        viewlib.display_add_button(self.w, links, self._cw.__("add"))

    def display_subobjects_list(self, entity):
        rset = entity.related(self.rtype, self.role)
        if rset:
            subvid = "seda.listitem"
            self._cw.view(
                "list",
                rset=rset,
                w=self.w,
                subvid=subvid,
                parent=self.parent(entity),
                tabid=self.__regid__,
            )


class SimplifiedArchiveUnitArchiveUnitsTab(ArchiveUnitArchiveUnitsTab):
    __select__ = ArchiveUnitArchiveUnitsTab.__select__ & simplified_profile()

    unit_custom_arguments = [
        ("unit_type", "unit_content", _("archive unit (content)")),
    ]


# data objects tab #############################################################


class ArchiveUnitDataObjectReferencesTab(ArchiveUnitSubObjectsTab):
    """Tab for data object references of an archive unit"""

    __regid__ = "seda_data_objects_tab"

    rtype_role_targets = [("seda_data_object_reference", "object", None)]

    _(
        "creating SEDADataObjectReference (SEDADataObjectReference seda_data_object_reference "
        "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement %(linkto)s)"
    )


class SimplifiedArchiveUnitDataObjectReferencesTab(ArchiveUnitDataObjectReferencesTab):
    """Tab for data object references of an archive unit within a simplified transfer: see
    referenced data object as if they were children of the archive unit referencing them (only one
    reference is allowed in such case, so this is fine).
    """

    __select__ = ArchiveUnitDataObjectReferencesTab.__select__ & simplified_profile()

    rtype_role_targets = [("seda_binary_data_object", "object", "SEDABinaryDataObject")]

    def display_add_button(self, entity):
        rtype_roles = [(rtype, role) for rtype, role, _ in self.rtype_role_targets]
        params = self.url_params(entity)
        params["referenced_by"] = str(entity.eid)
        transfer = entity.container[0]
        links = []
        if transfer.cw_etype == "SEDAArchiveTransfer":
            links = viewlib.add_links_from_rtype_roles(transfer, rtype_roles, params)
            label = self._cw.__("add")
        else:
            if has_rel_perm("add", entity, "seda_data_object_reference", "object"):
                vreg = self._cw.vreg
                url = (
                    vreg["etypes"]
                    .etype_class("SEDABinaryDataObject")
                    .cw_create_url(self._cw, **params)
                )
                label = self._cw.__("seda_binary_data_object_object")
                links = [
                    tags.a(
                        label, href=url, title=self._cw.__("New SEDABinaryDataObject")
                    )
                ]
        viewlib.display_add_button(self.w, links, self._cw.__("add"))

    def display_subobjects_list(self, entity):
        rset = self._cw.execute(
            "Any DO, DOUA, DOUC ORDERBY DOUA WHERE "
            "DO user_annotation DOUA, DO user_cardinality DOUC, "
            "REF seda_data_object_reference_id DO, REF seda_data_object_reference SEQ, "
            "SEQ eid %(x)s",
            {"x": entity.eid},
        )
        if rset:
            self._cw.view(
                "list",
                rset=rset,
                parent=self.parent(entity),
                w=self.w,
                subvid=self.subvid,
                tabid=self.tabid,
            )


# reference / content archive unit forms #######################################

# Top level ArchiveUnit form: create to distinct forms, one form archive unit reference and the
# other for archive unit content. This is done by a mix of uicfg, form and renderer customization
# depending on a 'unit_type' parameter in form params.

pvs.tag_object_of(("*", "seda_data_object_reference", CONTENT_ETYPE), "hidden")
afs.tag_object_of(("*", "seda_data_object_reference", CONTENT_ETYPE), "main", "hidden")

# add rtags for content archive unit in default afs
afs.tag_object_of(
    ("*", "seda_archive_unit_ref_id_from", "SEDAAltArchiveUnitArchiveUnitRefId"),
    "inlined",
    "hidden",
)
afs.tag_subject_of(
    (
        "SEDAAltArchiveUnitArchiveUnitRefId",
        "seda_seq_alt_archive_unit_archive_unit_ref_id_management",
        "*",
    ),
    "inlined",
    "inlined",
)

# and create a custom one for reference archive unit
au_ref_afs = afs.derive(
    __name__, is_instance("SEDAAltArchiveUnitArchiveUnitRefId") & is_archive_unit_ref()
)
au_ref_afs.tag_object_of(
    ("*", "seda_archive_unit_ref_id_from", "SEDAAltArchiveUnitArchiveUnitRefId"),
    "inlined",
    "inlined",
)
au_ref_afs.tag_subject_of(
    (
        "SEDAAltArchiveUnitArchiveUnitRefId",
        "seda_seq_alt_archive_unit_archive_unit_ref_id_management",
        "*",
    ),
    "inlined",
    "hidden",
)


class AltArchiveUnitRefIdSimplifiedAutomaticEntityForm(
    widgets.SimplifiedAutomaticEntityForm
):
    """Force display of AltArchiveUnitArchiveUnitRefId's sub-ArchiveUnitRefId or
    SeqAltArchiveUnitArchiveUnitRefIdManagement (uicfg will control which of them is displayed).
    """

    __select__ = widgets.SimplifiedAutomaticEntityForm.__select__ & is_instance(
        "SEDAAltArchiveUnitArchiveUnitRefId"
    )


class ArchiveUnitNoTitleEntityInlinedFormRenderer(
    widgets.NoTitleEntityInlinedFormRenderer
):
    """Don't display any title nor remove link for AltArchiveUnitArchiveUnitRefId,
    SeqAltArchiveUnitArchiveUnitRefIdManagement or ArchiveUnitRefId in the context of an archive
    unit reference.
    """

    __select__ = widgets.NoTitleEntityInlinedFormRenderer.__select__ & (
        is_instance("SEDAAltArchiveUnitArchiveUnitRefId", CONTENT_ETYPE)
        | (is_instance("SEDAArchiveUnitRefId") & is_archive_unit_ref())
    )


copy_afs = afs.derive(__name__, afs.__select__ & match_form_params(vid="copy"))
copy_afs.tag_subject_of(
    ("*", "seda_alt_archive_unit_archive_unit_ref_id", "*"), "main", "hidden"
)



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda unit tests for schema"""

from json import dumps
import unittest


from cubicweb.devtools.testlib import CubicWebTC
from cubicweb_web import INTERNAL_FIELD_VALUE

from cubicweb_seda.xsd2yams import RULE_TYPES
from cubicweb_seda.views import archiveunit, clone, export, mgmt_rules

from cubicweb_seda import testutils


def entity_fields(fields):
    return [
        str(field.name)
        for field in fields
        if field.eidparam and not field.name.startswith("__")
    ]


class ManagementRulesTC(CubicWebTC):
    def test_rule_ref_vocabulary(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            access_scheme = create("ConceptScheme", title="access")
            access_concept = access_scheme.add_concept(label="anyone")
            cnx.commit()

            bdo = testutils.create_transfer_to_bdo(cnx)
            transfer = bdo.container[0]

            rule_base = create("SEDAAccessRule", seda_access_rule=transfer)
            rule_alt = create(
                "SEDAAltAccessRulePreventInheritance",
                reverse_seda_alt_access_rule_prevent_inheritance=rule_base,
            )
            cnx.create_entity(
                "SEDAPreventInheritance", seda_prevent_inheritance=rule_alt
            )
            cnx.commit()

            self.assertEqual(
                mgmt_rules._rule_ref_vocabulary(transfer, rule_base.cw_etype),
                [
                    (
                        "you must specify a scheme for seda_access_rule_code_list_version_"
                        "from_object to select a value",
                        INTERNAL_FIELD_VALUE,
                    )
                ],
            )
            self.assertEqual(
                mgmt_rules._rule_ref_vocabulary(transfer, rule_alt.cw_etype),
                [
                    (
                        "you must specify a scheme for seda_access_rule_code_list_version_"
                        "from_object to select a value",
                        INTERNAL_FIELD_VALUE,
                    )
                ],
            )

            create(
                "SEDAAccessRuleCodeListVersion",
                seda_access_rule_code_list_version_from=transfer,
                seda_access_rule_code_list_version_to=access_scheme,
            )
            cnx.commit()

            self.assertEqual(
                mgmt_rules._rule_ref_vocabulary(transfer, rule_base.cw_etype),
                [
                    ("<no value specified>", "__cubicweb_internal_field__"),
                    (access_concept.label(), str(access_concept.eid)),
                ],
            )
            self.assertEqual(
                mgmt_rules._rule_ref_vocabulary(transfer, rule_alt.cw_etype),
                [
                    ("<no value specified>", "__cubicweb_internal_field__"),
                    (access_concept.label(), str(access_concept.eid)),
                ],
            )

    def test_archive_unit_component_rule_ref_vocabulary(self):
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)
            for rule_type in ("access", "appraisal"):
                etype = f"SEDASeq{rule_type.capitalize()}RuleRule"
                scheme = testutils.scheme_for_type(cnx, "seda_rule", etype)
                concept = scheme.add_concept(label="whatever")
                cnx.commit()

                unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                    None, cnx=cnx
                )

                self.assertEqual(
                    mgmt_rules._rule_ref_vocabulary(unit_alt_seq, etype),
                    [
                        ("<no value specified>", "__cubicweb_internal_field__"),
                        ("whatever", str(concept.eid)),
                    ],
                )


class PermissionsTC(CubicWebTC):
    """Functional test case about permissions in the web interface."""

    def test_container_permissions(self):
        """Check that a user cannot edit a SEDA profile he/she did not create."""
        alice_login = "alice"
        bob_login = "bob"
        with self.admin_access.cnx() as cnx:
            self.create_user(cnx, login=alice_login)
            self.create_user(cnx, login=bob_login)
            cnx.commit()
        with self.new_access(alice_login).web_request() as req:
            alice_profile = req.create_entity(
                "SEDAArchiveTransfer", title="Alice Profile"
            )
            req.cnx.commit()
        with self.new_access(bob_login).web_request() as req:
            bob_profile = req.create_entity("SEDAArchiveTransfer", title="Bob Profile")
            req.cnx.commit()
        with self.new_access(alice_login).web_request() as req:
            alice_profile = req.entity_from_eid(alice_profile.eid)
            bob_profile = req.entity_from_eid(bob_profile.eid)
            alice_profile_form = self.vreg["forms"].select(
                "edition", req, entity=alice_profile
            )
            bob_profile_form = self.vreg["forms"].select(
                "edition", req, entity=bob_profile
            )
            alice_field_names = [field.name for field in alice_profile_form.fields]
            bob_field_names = [field.name for field in bob_profile_form.fields]
            for field_name in (
                "title",
                "user_annotation",
                "seda_archival_agency",
                "seda_transferring_agency",
            ):
                self.assertIn(field_name, alice_field_names)
                self.assertNotIn(field_name, bob_field_names)


class RelationWidgetTC(CubicWebTC):
    """Functional test case about the relation widget."""

    def test_linkable_rset(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="Test widget")
            bdo = cnx.create_entity(
                "SEDABinaryDataObject",
                user_annotation="I am mandatory",
                seda_binary_data_object=transfer,
            )
            bdo_alt = cnx.create_entity(
                "SEDAAltBinaryDataObjectAttachment",
                reverse_seda_alt_binary_data_object_attachment=bdo,
            )
            cnx.create_entity("SEDAAttachment", seda_attachment=bdo_alt)
            compressed = cnx.create_entity("SEDACompressed", seda_compressed=bdo)
            cnx.commit()
        with self.admin_access.web_request() as req:
            compressed = req.entity_from_eid(compressed.eid)
            req.form = {
                "relation": "seda_algorithm:Concept:subject",
                "container": str(transfer.eid),
            }
            view = self.vreg["views"].select(
                "search_related_entities", req, rset=compressed.as_rset()
            )
            self.assertFalse(view.linkable_rset())
        with self.admin_access.cnx() as cnx:
            scheme = cnx.create_entity("ConceptScheme", title="CompressionAlgorithm")
            scheme.add_concept(label="bz2")
            cnx.create_entity(
                "SEDACompressionAlgorithmCodeListVersion",
                seda_compression_algorithm_code_list_version_from=transfer,
                seda_compression_algorithm_code_list_version_to=scheme,
            )
            cnx.commit()
        with self.admin_access.web_request() as req:
            compressed = req.entity_from_eid(compressed.eid)
            req.form = {
                "relation": "seda_algorithm:Concept:subject",
                "container": str(transfer.eid),
            }
            view = self.vreg["views"].select(
                "search_related_entities", req, rset=compressed.as_rset()
            )
            self.assertEqual(len(view.linkable_rset()), 1)


class HelperFunctionsTC(CubicWebTC):
    def test_rtags_from_xsd_element(self):
        from cubicweb_web.views.uicfg import reledit_ctrl
        from cubicweb_seda.views import rtags_from_xsd_element

        rsection, display_ctrl = rtags_from_xsd_element(
            "SEDABinaryDataObject", "FileInfo"
        )

        self.assertEqual(
            rsection.etype_get("SEDABinaryDataObject", "filename", "subject"), None
        )
        self.assertEqual(
            reledit_ctrl.etype_get("SEDABinaryDataObject", "filename", "subject"),
            {"novalue_include_rtype": False, "novalue_label": "<no value specified>"},
        )


class FormatSupportedPredicateTC(CubicWebTC):
    def test_format_supported(self):
        class fakecls:
            def __init__(self, seda_version):
                self.seda_version = seda_version

        predicate = export.format_supported()

        with self.admin_access.web_request() as req:
            transfer = req.create_entity(
                "SEDAArchiveTransfer", title="Test widget", simplified_profile=True
            )
            req.cnx.commit()

            self.assertEqual(
                predicate(fakecls("2.0"), req, entity=transfer), 0
            )  # not supported

            for seda_version in ("2.1", "1.0", "0.2"):
                self.assertEqual(
                    predicate(fakecls(seda_version), req, entity=transfer), 1
                )
            self.assertEqual(
                predicate(fakecls(seda_version="1000"), req, entity=transfer), 0
            )

            with req.cnx.security_enabled(write=False):
                transfer.cw_set(compat_list="SEDA 2.0")
                req.cnx.commit()
            transfer.cw_clear_all_caches()

            self.assertEqual(
                predicate(fakecls(seda_version="2.0"), req, entity=transfer), 1
            )

            req.form["version"] = "2.0"  # not considered
            self.assertEqual(
                predicate(fakecls(seda_version="1.0"), req, entity=transfer), 0
            )

            req.form["version"] = "2.0"
            self.assertEqual(predicate(None, req, entity=transfer), 1)

            req.form["version"] = "1.0"
            self.assertEqual(predicate(None, req, entity=transfer), 0)

            del req.form["version"]
            self.assertEqual(
                predicate(
                    fakecls(seda_version="1.0"), req, entity=transfer, version="2.0"
                ),
                0,
            )
            self.assertEqual(predicate(None, req, entity=transfer, version="2.0"), 1)
            self.assertEqual(predicate(None, req, entity=transfer, version="1.0"), 0)

            transfer.cw_set(simplified_profile=False)
            req.cnx.commit()  # will reset compat_list

            self.assertEqual(predicate(None, req, entity=transfer, version="2.0"), 0)
            self.assertEqual(predicate(None, req, entity=transfer, version="1.0"), 0)


class ArchiveTransferDiagnoseTabTC(CubicWebTC):
    def test_diagnose_tab(self):
        with self.admin_access.web_request() as req:
            transfer = req.create_entity(
                "SEDAArchiveTransfer", title="diagnosis testing"
            )
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            req.cnx.commit()
            # ensure the diagnosis tab display correctly
            self.view("seda_at_diagnose_tab", req=req, rset=transfer.as_rset())


class ArchiveTransferExportTC(CubicWebTC):
    def test_selected_export_class(self):
        with self.admin_access.web_request() as req:
            transfer = req.create_entity(
                "SEDAArchiveTransfer",
                title="diagnosis testing",
                simplified_profile=True,
            )
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            req.cnx.commit()

            exporter = self.vreg["views"].select("seda.export", req, entity=transfer)
            self.assertEqual(exporter.__class__, export.SEDAExportView)

            req.form["version"] = "0.2"
            exporter = self.vreg["views"].select("seda.export", req, entity=transfer)
            self.assertEqual(exporter.__class__, export.SEDAExportView)

            req.form["version"] = "3.0"
            exporter = self.vreg["views"].select("seda.export", req, entity=transfer)
            self.assertEqual(exporter.__class__, export.SEDANonSupportedExportView)

            with req.cnx.security_enabled(write=False):
                transfer.cw_set(compat_list="SEDA 2.0")
                req.cnx.commit()

            req.form["version"] = "0.2"
            exporter = self.vreg["views"].select("seda.export", req, entity=transfer)
            self.assertEqual(exporter.__class__, export.SEDANonSupportedExportView)

            req.form["version"] = "2.0"
            exporter = self.vreg["views"].select("seda.export", req, entity=transfer)
            self.assertEqual(exporter.__class__, export.SEDAExportView)

    def test_export_filename(self):
        with self.admin_access.web_request() as req:
            transfer = req.create_entity(
                "SEDAArchiveTransfer",
                title="diagnosis testing",
                simplified_profile=True,
            )
            req.cnx.commit()
            for version, fmt, expected_filename in (
                ("2.2", "rng", "diagnosis testing-2.2.rng"),
                ("2.2", "html", "diagnosis testing-2.2.html"),
                ("2.1", "rng", "diagnosis testing-2.1.rng"),
                ("1.0", "rng", "diagnosis testing-1.0.rng"),
                ("1.0", "xsd", "diagnosis testing-1.0.xsd"),
            ):
                req.form["version"], req.form["format"] = version, fmt
                view = self.vreg["views"].select(
                    "seda.export", req, rset=transfer.as_rset()
                )
                view.set_request_content_type()
                self.assertEqual(
                    view._cw.headers_out.getRawHeaders("content-disposition"),
                    [f'attachment;filename="{expected_filename}"'],
                )


class SimplifiedFormsTC(CubicWebTC):
    """Functional test case about forms in the web interface."""

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity(
                "SEDAArchiveTransfer",
                title="diagnosis testing",
                simplified_profile=True,
            )
            archive_unit = testutils.create_archive_unit(transfer)[0]

            cnx.commit()
            self.transfer_eid = transfer.eid
            self.archive_unit_eid = archive_unit.eid

    def create_and_link_form(self, req, etype, linkto=None, **kwargs):
        # add minimal information in form to mimick __linkto behaviour
        if linkto is None:
            linkto = self.archive_unit_eid
        req.form["__linkto"] = f"x:{linkto}:y"
        return self._create_etype_form(req, etype, **kwargs)

    def create_inlined_form(self, req, etype, parent=None, **kwargs):
        # add minimal information in form to mimick inlined form behaviour
        if parent is None:
            parent = self.archive_unit_eid
        req.form["arg"] = [dumps(str(parent))]
        return self._create_etype_form(req, etype, **kwargs)

    def _create_etype_form(self, req, etype, **kwargs):
        entity = req.vreg["etypes"].etype_class(etype)(req)
        return self.vreg["forms"].select("edition", req, entity=entity, **kwargs)

    def assertInlinedFields(self, form, expected):
        """Assert the given inlined form as the expected set of inlined forms, each defined by its
        field name and associated view class'name.
        """

        # str around field.name because it may be a relation schema instead of a string
        inlined_fields = [
            (str(field.name), field.view.__class__.__name__)
            for field in form.fields
            if hasattr(field, "view")
        ]
        self.assertEqual(set(inlined_fields), set(expected))

    def assertNoRemove(self, form, field_name, role):
        """Assert the given inlined form field will be rendered without a link to remove it."""
        field = form.field_by_name(field_name, role=role)
        self.assertEqual(field.view._get_removejs(), None)

    def assertNoTitle(self, form, field_name, role):
        """Assert the given inlined form field will be rendered without form title."""
        field = form.field_by_name(field_name, role=role)
        self.assertEqual(field.view.form_renderer_id, "notitle")

    def test_top_level_rule_form(self):
        with self.admin_access.web_request() as req:
            req.entity_from_eid(self.transfer_eid).cw_set(simplified_profile=False)
            req.cnx.commit()
            for rule_type in RULE_TYPES:
                if rule_type in ("classification", "hold"):
                    continue
                with self.subTest(rule_type=rule_type):
                    form = self.create_and_link_form(
                        req,
                        f"SEDA{rule_type.capitalize()}Rule",
                        linkto=self.transfer_eid,
                    )
                    self.assertInlinedFields(
                        form,
                        [
                            (
                                f"seda_seq_{rule_type}_rule_rule",
                                "InlineAddNewLinkView",
                            ),
                        ],
                    )
            form = self.create_and_link_form(
                req, "SEDAClassificationRule", linkto=self.transfer_eid
            )
            self.assertInlinedFields(
                form,
                [
                    ("seda_seq_classification_rule_rule", "InlineAddNewLinkView"),
                    ("seda_classification_audience", "InlineAddNewLinkView"),
                    ("seda_classification_reassessing_date", "InlineAddNewLinkView"),
                    ("seda_need_reassessing_authorization", "InlineAddNewLinkView"),
                ],
            )
            form = self.create_and_link_form(
                req, "SEDAHoldRule", linkto=self.transfer_eid
            )
            self.assertInlinedFields(
                form,
                [
                    (
                        "seda_seq_hold_rule_rule",
                        "SEDAWarningExportInlineAddNewLinkView",
                    ),
                ],
            )

    def test_top_level_rule_form_simplified(self):
        with self.admin_access.web_request() as req:
            for rule_type in ("appraisal", "access"):
                with self.subTest(rule_type=rule_type):
                    form = self.create_and_link_form(
                        req,
                        f"SEDA{rule_type.capitalize()}Rule",
                        linkto=self.transfer_eid,
                    )
                    self.assertInlinedFields(
                        form,
                        [
                            (
                                f"seda_seq_{rule_type}_rule_rule",
                                "RuleRuleInlineEntityCreationFormView",
                            ),
                        ],
                    )

    def test_rule_form(self):
        with self.admin_access.web_request() as req:
            for rule_type in ("access", "appraisal"):
                form = self.create_and_link_form(
                    req, f"SEDA{rule_type.capitalize()}Rule"
                )
                self.assertInlinedFields(
                    form,
                    [
                        (
                            f"seda_seq_{rule_type}_rule_rule",
                            "RuleRuleInlineEntityCreationFormView",
                        ),
                    ],
                )
                self.assertNoRemove(form, f"seda_seq_{rule_type}_rule_rule", "subject")

    def test_rule_rule_form(self):
        with self.admin_access.web_request() as req:
            for rule_type in ("access", "appraisal"):
                form = self.create_and_link_form(
                    req, f"SEDASeq{rule_type.capitalize()}RuleRule"
                )
                other_fields = [
                    str(field.name)
                    for field in form.fields
                    if not hasattr(field, "view")
                ]
                self.assertNotIn("user_cardinality", other_fields)
                self.assertInlinedFields(
                    form,
                    [
                        ("seda_start_date", "StartDateInlineEntityCreationFormView"),
                    ],
                )
                self.assertNoRemove(form, "seda_start_date", "object")
                self.assertNoTitle(form, "seda_start_date", "object")

    def test_create_data_object_full(self):
        with self.admin_access.web_request() as req:
            req.entity_from_eid(self.transfer_eid).cw_set(simplified_profile=False)
            req.cnx.commit()
            form = self.create_inlined_form(
                req, "SEDADataObjectReference", formtype="inlined"
            )
            self.assertEqual(
                entity_fields(form.fields), ["seda_data_object_reference_id"]
            )

    def test_create_data_object_simplified(self):
        with self.admin_access.web_request() as req:
            form = self.create_inlined_form(
                req, "SEDADataObjectReference", formtype="inlined"
            )
            self.assertEqual(entity_fields(form.fields), [])

    def test_create_data_object_from_archive_unit_component(self):
        with self.admin_access.web_request() as req:
            unit = testutils.create_archive_unit(None, cnx=req.cnx)[0]
            req.cnx.commit()
            form = self.create_inlined_form(
                req, "SEDADataObjectReference", formtype="inlined", parent=unit.eid
            )
            self.assertEqual(entity_fields(form.fields), [])


class CloneActionsTC(CubicWebTC):
    def test_archive_transfer(self):
        with self.admin_access.web_request() as req:
            transfer = req.create_entity("SEDAArchiveTransfer", title="test")
            unit = testutils.create_archive_unit(transfer)[0]
            req.cnx.commit()

            actions = self.pactionsdict(req, transfer.as_rset())
            self.assertIn(clone.SEDAArchiveTransferCloneAction, actions["mainactions"])

            actions = self.pactionsdict(req, unit.as_rset())
            self.assertNotIn(clone.SEDAArchiveUnitCloneAction, actions["mainactions"])

    def test_unit_component(self):
        with self.admin_access.web_request() as req:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=req)
            subunit = testutils.create_archive_unit(unit_alt_seq)[0]
            req.cnx.commit()

            actions = self.pactionsdict(req, unit.as_rset())
            self.assertIn(clone.SEDAArchiveUnitCloneAction, actions["mainactions"])

            actions = self.pactionsdict(req, subunit.as_rset())
            self.assertNotIn(clone.SEDAArchiveUnitCloneAction, actions["mainactions"])


class CloneImportTC(CubicWebTC):
    """Tests for 'seda.doimport' controller (called from JavaScript)."""

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test")
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                None, cnx=cnx, user_cardinality="1", user_annotation="plop"
            )
            bdo = testutils.create_data_object(unit_alt_seq, filename="file.txt")
            self.transfer_eid = transfer.eid
            self.unit_eid = unit.eid
            self.bdo_eid = bdo.eid
            cnx.commit()

    def doctypecode(self, cnx):
        return cnx.create_entity(
            "SEDADocumentTypeCode",
            seda_document_type_code_value=self.doctypecodevalue_eid,
        )

    def test_import_one_entity(self):
        params = dict(eid=str(self.transfer_eid), cloned=str(self.unit_eid))
        with self.admin_access.web_request(**params) as req:
            path, _ = self.expect_redirect_handle_request(req, "seda.doimport")
            etype, eid = path.split("/")
            self.assertEqual(etype, "SEDAArchiveTransfer".lower())
            clone = req.execute(
                "Any X WHERE X seda_archive_unit P, P eid %(p)s", {"p": eid}
            ).one()
            self.assertEqual([x.eid for x in clone.clone_of], [self.unit_eid])
            # Check that original entity attributes have been copied.
            self.assertEqual(clone.user_cardinality, "1")
            self.assertEqual(clone.user_annotation, "plop")
            # Check that data object has been copied and linked to the transfer
            seq = clone.first_level_choice.content_sequence
            bdo = seq.reverse_seda_data_object_reference[
                0
            ].seda_data_object_reference_id[0]
            self.assertNotEqual(bdo.eid, self.bdo_eid)
            self.assertEqual(bdo.filename, "file.txt")
            transfer = req.entity_from_eid(self.transfer_eid)
            transfer_bdos = [do.eid for do in transfer.reverse_seda_binary_data_object]
            self.assertEqual(transfer_bdos, [bdo.eid])

    def test_import_multiple_entities(self):
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                None, cnx=cnx, user_cardinality="0..1", user_annotation="plouf"
            )
            cnx.commit()
        to_clone = [self.unit_eid, unit.eid]
        params = dict(
            eid=str(self.transfer_eid),
            cloned=",".join([str(self.unit_eid), str(unit.eid)]),
        )
        with self.admin_access.web_request(**params) as req:
            path, _ = self.expect_redirect_handle_request(req, "seda.doimport")
            etype, eid = path.split("/")
            self.assertEqual(etype, "SEDAArchiveTransfer".lower())
            rset = req.execute(
                "Any X,O WHERE X seda_archive_unit P, P eid %(p)s, X clone_of O",
                {"p": eid},
            )
            self.assertEqual(len(rset), 2)
            self.assertCountEqual([oeid for __, oeid in rset.rows], to_clone)
            cardinalities, annotations = list(
                zip(
                    *[
                        (clone.user_cardinality, clone.user_annotation)
                        for clone in rset.entities()
                    ]
                )
            )
            self.assertCountEqual(cardinalities, ("1", "0..1"))
            self.assertCountEqual(annotations, ("plop", "plouf"))


class SEDATreeTC(CubicWebTC):
    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            self.transfer_eid = cnx.create_entity(
                "SEDAArchiveTransfer", title="Test", simplified_profile=True
            ).eid
            cnx.commit()

    def test_archiveunit_reparent_to_transfer(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            unit2 = testutils.create_archive_unit(unit_alt_seq)[0]
            cnx.commit()
            unit2.cw_adapt_to("IJQTree").reparent(transfer.eid, 0)
            cnx.commit()
            transfer.cw_clear_all_caches()
            self.assertEqual(
                [x.eid for x in transfer.reverse_seda_archive_unit],
                [unit2.eid, unit.eid],
            )

    def test_archiveunit_reparent_to_archiveunit(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            unit, _, unit_alt_seq = testutils.create_archive_unit(transfer)
            unit2, _, unit2_alt_seq = testutils.create_archive_unit(transfer)
            subunit, _, _ = testutils.create_archive_unit(unit2_alt_seq)
            cnx.commit()
            unit.cw_clear_all_caches()
            unit.cw_adapt_to("IJQTree").reparent(unit2.eid, 0)
            cnx.commit()
            unit2_alt_seq.cw_clear_all_caches()
            self.assertEqual(
                [x.eid for x in unit2_alt_seq.reverse_seda_archive_unit],
                [unit.eid, subunit.eid],
            )

    def test_binarydataobject_reparent(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            unit, _, unit_alt_seq = testutils.create_archive_unit(transfer)
            unit2, _, unit2_alt_seq = testutils.create_archive_unit(transfer)
            bdo = testutils.create_data_object(transfer)
            ref = cnx.create_entity(
                "SEDADataObjectReference",
                seda_data_object_reference=unit_alt_seq,
                seda_data_object_reference_id=bdo,
            )
            cnx.commit()
            bdo.cw_clear_all_caches()
            bdo.cw_adapt_to("IJQTree").reparent(unit2.eid, 0)
            cnx.commit()
            unit2_alt_seq.cw_clear_all_caches()
            self.assertEqual(
                [x.eid for x in unit2_alt_seq.reverse_seda_data_object_reference],
                [ref.eid],
            )
            unit_alt_seq.cw_clear_all_caches()
            self.assertFalse(unit_alt_seq.reverse_seda_data_object_reference)

    def test_binarydataobject_reorder_full(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            transfer.cw_set(simplified_profile=False)
            bdo1 = testutils.create_data_object(transfer)
            bdo2 = testutils.create_data_object(transfer)
            cnx.commit()

            transfer.cw_adapt_to("IJQTree").move_child_at_index(bdo2.eid, 0)

            transfer.cw_clear_all_caches()
            self.assertEqual(
                [x.eid for x in transfer.reverse_seda_binary_data_object],
                [bdo2.eid, bdo1.eid],
            )

            bdo2.cw_clear_all_caches()
            transfer.cw_adapt_to("IJQTree").move_child_at_index(bdo2.eid, 2)

            bdo1.cw_clear_all_caches()
            bdo2.cw_clear_all_caches()

            transfer.cw_clear_all_caches()
            self.assertEqual(
                [x.eid for x in transfer.cw_adapt_to("ITreeBase").iterchildren()],
                [bdo1.eid, bdo2.eid],
            )

    def test_binarydataobject_reorder_simplified(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            unit, _, unit_alt_seq = testutils.create_archive_unit(transfer)
            bdo1 = testutils.create_data_object(unit_alt_seq)
            bdo2 = testutils.create_data_object(unit_alt_seq)
            cnx.commit()

            unit.cw_adapt_to("IJQTree").move_child_at_index(bdo2.eid, 0)

            transfer.cw_clear_all_caches()
            self.assertEqual(
                [x.eid for x in unit.cw_adapt_to("ITreeBase").iterchildren()],
                [bdo2.eid, bdo1.eid],
            )


class FakeForm:
    def __init__(self, req, edited_entity):
        self._cw = req
        self.edited_entity = edited_entity


class ArchiveUnitVocabularyTC(CubicWebTC):
    def test_unit_ref_vocabulary(self):
        with self.admin_access.web_request() as req:
            transfer = req.cnx.create_entity("SEDAArchiveTransfer", title="Test")
            archunit, _, alt_seq = testutils.create_archive_unit(transfer)
            req.cnx.commit()
            archunit.cw_clear_all_caches()
            # actually expect a archive unit reference, but we want to test the query so any object
            # linked to the container is fine
            form = FakeForm(req, archunit)
            self.assertEqual(
                archiveunit.unit_ref_vocabulary(form, None),
                [("archive unit title", str(archunit.eid))],
            )

    def test_do_ref_vocabulary(self):
        with self.admin_access.web_request() as req:
            transfer = req.cnx.create_entity("SEDAArchiveTransfer", title="Test")
            bdo = testutils.create_data_object(transfer)
            req.cnx.commit()
            bdo.cw_clear_all_caches()
            # actually expect a data object reference, but we want to test the query so any object
            # linked to the container is fine
            form = FakeForm(req, bdo)
            self.assertEqual(
                archiveunit.do_ref_vocabulary(form, None),
                [("data object title", str(bdo.eid))],
            )


class SEDANavigationTC(CubicWebTC):
    def test_breadcrumbs(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            unit1 = testutils.create_archive_unit(transfer)[0]
            unit2 = testutils.create_archive_unit(None, cnx=cnx)[0]
            cnx.commit()
        with self.admin_access.web_request() as req:
            unit1 = req.entity_from_eid(unit1.eid)
            unit2 = req.entity_from_eid(unit2.eid)
            # unit1 is related to a transfer
            breadcrumbs = unit1.cw_adapt_to("IBreadCrumbs").breadcrumbs()
            expected_breadcrumbs = [transfer, unit1]
            self.assertEqual(breadcrumbs, expected_breadcrumbs)
            # unit2 is not related to a transfer, breadcrumbs leads to /sedalib.
            breadcrumbs = unit2.cw_adapt_to("IBreadCrumbs").breadcrumbs()
            expected_breadcrumbs = [
                ("http://testing.cubicweb/cubicweb/sedalib", "SEDAArchiveUnit_plural"),
                unit2,
            ]
            self.assertEqual(breadcrumbs, expected_breadcrumbs)


if __name__ == "__main__":
    unittest.main()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda tools to diagnose versions compatibility of profiles."""

from collections import namedtuple
from itertools import chain

from cubicweb import _
from cubicweb_web.view import EntityAdapter
from cubicweb.predicates import is_instance


ALL_FORMATS = frozenset(
    ("SEDA 2.2", "SEDA 2.1", "SEDA 1.0", "SEDA 0.2", "simplified", "RNG")
)

Rule = namedtuple("Rule", ["impacted_formats", "message", "tab_id", "watch"])
RULES = {
    "use_archive_unit_ref": Rule(
        {"SEDA 1.0", "SEDA 0.2", "simplified"},
        _("Archive unit reference are only supported by SEDA 2."),
        "main_tab",
        {
            "seda_seq_access_rule_rule",
            "seda_seq_appraisal_rule_rule",
        },
    ),
    "seda1_need_access_rule": Rule(
        {"SEDA 1.0"},
        _(
            "First level archive unit must have an associated access rule to be exportable "
            "in SEDA 1. You may define it on the archive unit or as a default rule on the "
            "transfer."
        ),
        "seda_management_tab",
        {
            "seda_archive_unit",
            "seda_alt_archive_unit_archive_unit_ref_id",
            "seda_seq_alt_archive_unit_archive_unit_ref_id_management",
            "seda_access_rule",
        },
    ),
    "rule_without_rule": Rule(
        {"SEDA 1.0", "SEDA 0.2", "simplified"},
        _("Some management rule has no inner rule, one is required."),
        "seda_management_tab",
        {
            "seda_seq_access_rule_rule",
            "seda_seq_appraisal_rule_rule",
        },
    ),
    "rule_with_too_much_rules": Rule(
        {"SEDA 1.0", "SEDA 0.2", "simplified"},
        _(
            "Some management rule has more than one inner rules, a single one is required."
        ),
        "seda_management_tab",
        {
            "seda_seq_access_rule_rule",
            "seda_seq_appraisal_rule_rule",
        },
    ),
    "rule_unsupported_card": Rule(
        {"SEDA 1.0", "SEDA 0.2", "simplified"},
        _("Inner rule has cardinality other than 1."),
        "seda_management_tab",
        {
            ("SEDASeqAccessRuleRule", "user_cardinality"),
            ("SEDASeqAppraisalRuleRule", "user_cardinality"),
        },
    ),
    "rule_need_start_date": Rule(
        {"SEDA 1.0", "SEDA 0.2", "simplified"},
        _("Inner rule has no start date."),
        "seda_management_tab",
        {
            "seda_start_date",
        },
    ),
    "rule_start_unsupported_card": Rule(
        {"SEDA 1.0", "SEDA 0.2", "simplified"},
        _("Start date has cardinality other than 1."),
        "seda_management_tab",
        {
            ("SEDAStartDate", "user_cardinality"),
        },
    ),
    "rule_ref_non_rule_id": Rule(
        {"SEDA 1.0", "SEDA 0.2", "simplified"},
        _(
            "Rule has an explicit rule deactivation - only ignore all rules is supported in "
            "simplified profiles."
        ),
        "seda_management_tab",
        {
            "seda_ref_non_rule_id_from",
        },
    ),
    "seda02_custodial_history_items": Rule(
        {"SEDA 0.2"},
        _(
            "Custodial history is text with SEDA 0.2, hence only one item element is considered."
        ),
        "seda_history_tab",
        {
            "seda_custodial_history_item",
        },
    ),
    "seda02_custodial_history_when": Rule(
        {"SEDA 0.2"},
        _(
            "Custodial history is text with SEDA 0.2, hence date information isn't considered."
        ),
        "seda_history_tab",
        {
            "seda_when",
        },
    ),
    "rng_ambiguity": Rule(
        {"RNG"},
        _(
            "More than one children with cardinality not equal to '1', "
            "this may cause RNG validation problems."
        ),
        None,  # generated at run time
        [],
    ),  # have it's own hook
    "rng_mandatory_not_first": Rule(
        {"RNG"},
        _(
            "Children with cardinality equal to '1' should appear before other "
            "to avoid potential RNG validation problems."
        ),
        None,  # generated at run time
        [],
    ),  # have it's own hook
}


class CompatError(
    namedtuple(
        "_CompatError", ["impacted_formats", "message", "tab_id", "entity", "rule_id"]
    )
):
    """Convenience class holding information about a problem in a profile forbidding export to some
    format:

    * `impacted_formats`: set of formats that are no more available because of this error,
    * `message`: message string explaining the problem,
    * `entity`: 1st class entity where the error lies (one of archive unit, data object, etc.),
    * `tab`: entity's tab where the problem may be fixed.
    """

    def __new__(cls, rule_id, entity, **kwargs):
        rule = RULES[rule_id]
        impacted_formats = kwargs.get("impacted_formats", rule.impacted_formats)
        message = kwargs.get("message", rule.message)
        tab_id = kwargs.get("tab_id", rule.tab_id)
        return super().__new__(cls, impacted_formats, message, tab_id, entity, rule_id)


class ISEDACompatAnalyzer(EntityAdapter):
    """Adapter that will analyze profile to diagnose its format compatibility (SEDA 2, SEDA 1 and/or
    SEDA 0.2) while being to explain the problem to end-users.
    """

    __regid__ = "ISEDACompatAnalyzer"
    __select__ = is_instance("SEDAArchiveTransfer")

    def diagnose(self):
        possible_formats = set(ALL_FORMATS)
        for compat_error in self.detect_problems():
            possible_formats -= compat_error.impacted_formats
        return possible_formats

    def detect_problems(self):
        """Yield :class:`CompatError` describing a problem that prevents the given profile to be
        compatible with some format.
        """
        for args in self.failing_rules():
            assert len(args) in (2, 3)
            extra_kwargs = args[2] if len(args) == 3 else {}
            yield CompatError(*args[:2], **extra_kwargs)

    def failing_rules(self):
        """Yield (rule identifier, problematic entity) describing a problem that prevents the given
        profile to be compatible with some format.
        """
        for check_method in (
            self._check_usage_of_reference_archive_unit,
            self._check_management_rules,
            self._check_custodial_history_rules,
            self._check_rng_ambiguities,
        ):
            yield from check_method()

    def _check_management_rules(self):  # noqa (too complex)
        profile = self.entity
        # First level archive unit needs an access rule (SEDA 1)
        if not profile.reverse_seda_access_rule:
            for archive_unit in profile.archive_units:
                seq = archive_unit.first_level_choice.content_sequence
                if seq is None:
                    # reference archive unit
                    continue
                if not seq.reverse_seda_access_rule:
                    yield "seda1_need_access_rule", archive_unit
        # Access/appraisal rule must have one and only one sequence, which have one and only one
        # start date (and both of them must have 1 cardinality)
        for rule in self._cw.execute(
            "Any X WHERE X is IN (SEDAAccessRule, SEDAAppraisalRule), "
            "X container C, C eid %(c)s",
            {"c": profile.eid},
        ).entities():
            if not rule.rules:
                yield "rule_without_rule", _parent(rule)
            elif len(rule.rules) > 1:
                yield "rule_with_too_much_rules", _parent(rule)
            else:
                rule_seq = rule.rules[0]
                if rule_seq.user_cardinality != "1":
                    yield "rule_unsupported_card", _parent(rule)
                if not rule_seq.start_date:
                    yield "rule_need_start_date", _parent(rule)
                elif rule_seq.start_date.user_cardinality != "1":
                    yield "rule_start_unsupported_card", _parent(rule)
        # Access/appraisal rule shouldn't use ref non rule id
        for rule_type in ("access", "appraisal"):
            for rule in self._cw.execute(
                "DISTINCT Any X WHERE X seda_alt_{}_rule_prevent_inheritance ALT,"
                "REF seda_ref_non_rule_id_from ALT, "
                "X container C, C eid %(c)s".format(rule_type),
                {"c": profile.eid},
            ).entities():
                yield "rule_ref_non_rule_id", _parent(rule)

    def _check_custodial_history_rules(self):
        profile = self.entity
        # Check for more than one custodial history items (SEDA 0.2)
        for content in self._cw.execute(
            "Any CONT GROUPBY CONT WHERE X seda_custodial_history_item CONT,"
            "X container C, C eid %(c)s HAVING COUNT(X) > 1",
            {"c": profile.eid},
        ).entities():
            parent = _parent(content) if profile.simplified_profile else content
            yield "seda02_custodial_history_items", parent
        # Check for SEDAwhen usage (SEDA 0.2)
        for content in self._cw.execute(
            "DISTINCT Any CONT WHERE X seda_custodial_history_item CONT,"
            "W seda_when X, X container C, C eid %(c)s",
            {"c": profile.eid},
        ).entities():
            parent = _parent(content) if profile.simplified_profile else content
            yield "seda02_custodial_history_when", parent

    def _check_usage_of_reference_archive_unit(self):
        for archive_unit in self._cw.execute(
            "DISTINCT Any X WHERE X is SEDAArchiveUnit,"
            "X seda_alt_archive_unit_archive_unit_ref_id ALT, "
            "NOT EXISTS(ALT seda_seq_alt_archive_unit_archive_unit_ref_id_management SEQ),"
            "X container C, C eid %(c)s",
            {"c": self.entity.eid},
        ).entities():
            yield "use_archive_unit_ref", archive_unit

    def _check_rng_ambiguities(self):
        for rql, error_id in [
            (_AMBIGUOUS_RQL, "rng_ambiguity"),
            (_MANDATORY_NOT_LAST_AMBIGUOUS_RQL, "rng_mandatory_not_first"),
        ]:
            for eid, rtype in self._cw.execute(rql, {"c": self.entity.eid}):
                # if the container is a simplified profile, allow to have several
                # entities with cardinality != 1 under the seda_binary_data_object
                # relation because data objects are all linked to the transfer
                # through this relation, but this is only relevant for SEDA 2 export
                # and we don't want to prevent this in profiles for SEDA 0.2 and 1.0
                # export.
                if (
                    rtype == "seda_binary_data_object"
                    and self.entity.simplified_profile
                ):
                    continue
                tab_id = _RTYPE_TO_TAB.get(rtype)
                yield error_id, _parent(self._cw.entity_from_eid(eid)), {
                    "tab_id": tab_id
                }


def _parent(entity):
    """Return the first encountered parent which is an ArchiveUnit"""
    while entity.cw_etype not in (
        "SEDAArchiveTransfer",
        "SEDAArchiveUnit",
        "SEDABinaryDataObject",
        "SEDAPhysicalDataObject",
    ):
        entity = entity.cw_adapt_to("IContained").parent
    return entity


_RTYPE_TO_TAB = {
    "seda_addressee_from": "seda_agents_tab",
    "seda_archive_unit": "seda_archive_units_tab",
    "seda_binary_data_object": "seda_data_objects_tab",
    "seda_custodial_history_item": "seda_history_tab",
    "seda_data_object_reference": "seda_data_objects_tab",
    "seda_event": "seda_events_tab",
    "seda_is_part_of": "seda_relations_tab",
    "seda_is_version_of": "seda_relations_tab",
    "seda_juridictional": "seda_coverage_tab",
    "seda_keyword": "seda_indexation_tab",
    "seda_physical_data_object": "seda_data_objects_tab",
    "seda_recipient_from": "seda_agents_tab",
    # don't care about order on this one, for now at least
    # 'seda_ref_non_rule_id_from': 'seda_management_tab',
    "seda_references": "seda_relations_tab",
    "seda_related_transfer_reference": "seda_at_related_transfers_tab",
    "seda_relationship": "seda_do_relations",
    "seda_replaces": "seda_relations_tab",
    "seda_requires": "seda_relations_tab",
    "seda_spatial": "seda_coverage_tab",
    "seda_tag": "seda_indexation_tab",
    "seda_temporal": "seda_coverage_tab",
    "seda_writer_from": "seda_agents_tab",
}

_AMBIGUOUS_RQL = " UNION ".join(
    chain(
        (
            '(Any P, "{rtype}" GROUPBY P WHERE X {rtype} P, P container C, C eid %(c)s, '
            'NOT X user_cardinality "1" HAVING COUNT(X) > 1)'.format(rtype=rtype)
            for rtype in _RTYPE_TO_TAB
            if rtype
            not in (
                "seda_binary_data_object",
                "seda_physical_data_object",
                "seda_related_transfer_reference",
            )
        ),
        # we need another RQL for top level entities without container relation
        (
            '(Any P, "{rtype}" GROUPBY P WHERE X {rtype} P, P eid %(c)s, '
            'NOT X user_cardinality "1" HAVING COUNT(X) > 1)'.format(rtype=rtype)
            for rtype in (
                "seda_archive_unit",
                "seda_binary_data_object",
                "seda_physical_data_object",
                "seda_related_transfer_reference",
            )
        ),
    )
)

_MANDATORY_NOT_LAST_AMBIGUOUS_RQL = " UNION ".join(
    chain(
        (
            '(Any P, "{rtype}" WHERE EXISTS(X1 {rtype} P, X2 {rtype} P, '
            "P container C, C eid %(c)s, "
            'X1 user_cardinality "1", NOT X2 user_cardinality "1", '
            "X1 ordering ORD, X2 ordering < ORD))".format(rtype=rtype)
            for rtype in _RTYPE_TO_TAB
            if rtype
            not in (
                "seda_binary_data_object",
                "seda_physical_data_object",
                "seda_related_transfer_reference",
            )
        ),
        # we need another RQL for top level entities without container relation
        (
            '(Any P, "{rtype}" WHERE EXISTS(X1 {rtype} P, X2 {rtype} P, P eid %(c)s, '
            'X1 user_cardinality "1", NOT X2 user_cardinality "1", '
            "X1 ordering ORD, X2 ordering < ORD))".format(rtype=rtype)
            for rtype in (
                "seda_archive_unit",
                "seda_binary_data_object",
                "seda_physical_data_object",
                "seda_related_transfer_reference",
            )
        ),
    )
)



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Generate Yams schema from XSD file. XSD parsing is done using pyxst.
"""

try:
    from yams import BASE_TYPES
except ImportError:
    BASE_TYPES = {"String", "Boolean", "Int", "TZDatetime", "Date", "Bytes", "Decimal"}

from pyxst.xml_struct import graph_nodes

from cubicweb_seda.xsd import XSDMMapping, un_camel_case


EXT_ETYPES = {"AuthorityRecord", "ConceptScheme", "Concept"}
RULE_TYPES = {
    "access",
    "appraisal",
    "classification",
    "reuse",
    "dissemination",
    "storage",
    "hold",
}

# elements in the intermediary model but not in the yams model
SKIP_ETYPES = {"SEDAid", "SEDAhref", "SEDAfilename"}
SKIP_ATTRS = {
    "acquired_date",
    "birth_date",
    "classification_reassessing_date",
    "hold_reassessing_date",
    "hold_end_date",
    "created_date",
    "date",
    "date_created_by_application",
    "date_signature",
    "death_date",
    "depth",
    "diameter",
    "end_date",
    "event_date_time",
    "event_detail",
    "event_identifier",
    "gps_altitude",
    "gps_altitude_ref",
    "gps_date_stamp",
    "gps_latitude",
    "gps_latitude_ref",
    "gps_longitude",
    "gps_longitude_ref",
    "gps_version_id",
    "height",
    "href",
    "last_modified",
    "length",
    "message_identifier",
    "number_of_page",
    "physical_id",
    "received_date",
    "registered_date",
    "related_transfer_reference",
    "repository_archive_unit_pid",
    "repository_object_pid",
    "restriction_end_date",
    "originating_system_id",
    "sent_date",
    "shape",
    "start_date",
    "size",
    "system_id",
    "thickness",
    "transacted_date",
    "transfer_request_reply_identifier",
    "uncompressed_size",
    "uri",
    "weight",
    "when",
    "width",
}

# list of entity types that may be used multiple times at a same level, and
# through which relation
MULTIPLE_CHILDREN = [
    ("SEDAArchiveUnit", "seda_archive_unit"),
    ("SEDABinaryDataObject", "seda_binary_data_object"),
    ("SEDAPhysicalDataObject", "seda_physical_data_object"),
    ("SEDADataObjectReference", "seda_data_object_reference"),
    ("SEDARelatedTransferReference", "seda_related_transfer_reference"),
    ("SEDAWriter", "seda_writer_from"),
    ("SEDAAddressee", "seda_addressee_from"),
    ("SEDARecipient", "seda_recipient_from"),
    ("SEDASpatial", "seda_spatial"),
    ("SEDATemporal", "seda_temporal "),
    ("SEDAJuridictional", "seda_juridictional"),
    ("SEDAKeyword", "seda_keyword"),
    ("SEDATag", "seda_tag"),
    ("SEDAIsVersionOf", "seda_is_version_of"),
    ("SEDAReplaces", "seda_replaces"),
    ("SEDARequires", "seda_requires"),
    ("SEDAIsPartOf", "seda_is_part_of"),
    ("SEDAReferences", "seda_references"),
    ("SEDAEvent", "seda_event"),
    ("SEDACustodialHistoryItem", "seda_custodial_history_item"),
    ("SEDARelationship", "seda_relationship"),
]

RDEF_CONSTRAINTS = {
    # x-ref constraints
    "seda_data_object_reference_id": "S container C, O container C",
    "seda_archive_unit_ref_id_to": "S container C, O container C",
    "seda_signed_object_id": "S container C, O container C",
    "seda_target": "S container C, O container C",
    # link to concept constraints
    # - scheme defined according to the relation type
    "seda_description_level_to": (
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_description_level_to"'
    ),
    "seda_classification_level": (
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_classification_level"'
    ),
    "seda_language_to": (
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_language_to"'
    ),
    "seda_legal_status_to": (
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_legal_status_to"'
    ),
    "seda_description_language_to": (
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_description_language_to"'
    ),
    "seda_event_type_to": (
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_event_type_to"'
    ),
    "seda_keyword_type_to": (
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_keyword_type_to"'
    ),
    "seda_type_to": (
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_type_to"'
    ),
    # - scheme defined according to the relation type and the subject type
    "seda_algorithm": (  # XXX still there because of signature
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_algorithm", '
        'CS scheme_entity_type ET, ET name "{subjtype}"'
    ),
    "seda_unit": (
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_unit", '
        'CS scheme_entity_type ET, ET name "{subjtype}"'
    ),
    "seda_final_action": (
        'O in_scheme CS, CS scheme_relation_type CR, CR name "seda_final_action", '
        'CS scheme_entity_type ET, ET name "{subjtype}"'
    ),
    # - scheme defined as code list
    #   if you modify one of those, keep synchronized the SCHEME_FROM_CONTAINER map below
    #   and run "python -m tox -e make" to update the schema"
    "seda_acquisition_information_to": (
        "O in_scheme CS, "
        "EXISTS(CACLV seda_acquisition_information_code_list_version_from AT, "
        "       CACLV seda_acquisition_information_code_list_version_to CS,"
        "       S container AT)"
        " OR EXISTS(S container AU, AU is SEDAArchiveUnit, CS scheme_relation_type RT, "
        '           RT name "seda_acquisition_information_to")'
    ),
    "seda_mime_type_to": (
        "O in_scheme CS, "
        "EXISTS(CACLV seda_mime_type_code_list_version_from AT, "
        "       CACLV seda_mime_type_code_list_version_to CS,"
        "       S container AT)"
        " OR EXISTS(S container AU, AU is SEDAArchiveUnit, CS scheme_relation_type RT, "
        '           RT name "file_category")'
    ),
    "seda_encoding_to": (
        "O in_scheme CS, "
        "EXISTS(CACLV seda_encoding_code_list_version_from AT, "
        "       CACLV seda_encoding_code_list_version_to CS,"
        "       S container AT)"
        " OR EXISTS(S container AU, AU is SEDAArchiveUnit, CS scheme_relation_type RT, "
        '           RT name "seda_encoding_to")'
    ),
    "seda_format_id_to": (
        "O in_scheme CS, "
        "EXISTS(CACL seda_file_format_code_list_version_from AT, "
        "       CACL seda_file_format_code_list_version_to CS, "
        "       S container AT)"
        " OR EXISTS(S container AU, AU is SEDAArchiveUnit, CS scheme_relation_type RT, "
        '           RT name "file_category")'
    ),
    "seda_data_object_version_to": (
        "O in_scheme CS, CACLV seda_data_object_version_code_list_version_from AT, "
        "CACLV seda_data_object_version_code_list_version_to CS,"
        "S container AT"
    ),
    "seda_type_relationship": (
        "O in_scheme CS, CACLV seda_relationship_code_list_version_from AT, "
        "CACLV seda_relationship_code_list_version_to CS,"
        "S container AT"
    ),
    ("SEDABinaryDataObject", "seda_algorithm"): (
        "O in_scheme CS, "
        "EXISTS(S container AT, CACLV seda_message_digest_algorithm_code_list_version_from AT, "
        "       CACLV seda_message_digest_algorithm_code_list_version_to CS) "
        "OR EXISTS(S container AU, AU is SEDAArchiveUnit, "
        '          CS scheme_relation_type RT, RT name "seda_algorithm", '
        '          CS scheme_entity_type ET, ET name "SEDABinaryDataObject")'
    ),
    ("SEDACompressed", "seda_algorithm"): (
        "O in_scheme CS, CACLV seda_compression_algorithm_code_list_version_from AT, "
        "CACLV seda_compression_algorithm_code_list_version_to CS,"
        "S container AT"
    ),
    ("SEDASeqClassificationRuleRule", "seda_rule"): (
        "O in_scheme CS, CACLV seda_classification_rule_code_list_version_from AT, "
        "CACLV seda_classification_rule_code_list_version_to CS,"
        "S container AT"
    ),
    ("SEDASeqReuseRuleRule", "seda_rule"): (
        "O in_scheme CS, CACLV seda_reuse_rule_code_list_version_from AT, "
        "CACLV seda_reuse_rule_code_list_version_to CS,"
        "S container AT"
    ),
    ("SEDASeqDisseminationRuleRule", "seda_rule"): (
        "O in_scheme CS, CACLV seda_dissemination_rule_code_list_version_from AT, "
        "CACLV seda_dissemination_rule_code_list_version_to CS,"
        "S container AT"
    ),
    ("SEDASeqAccessRuleRule", "seda_rule"): (
        "O in_scheme CS, "
        "EXISTS(CACLV seda_access_rule_code_list_version_from AT, "
        "       CACLV seda_access_rule_code_list_version_to CS,"
        "       S container AT)"
        " OR EXISTS(S container AU, AU is SEDAArchiveUnit)"
    ),
    ("SEDASeqAppraisalRuleRule", "seda_rule"): (
        "O in_scheme CS, "
        "EXISTS(CACLV seda_appraisal_rule_code_list_version_from AT, "
        "       CACLV seda_appraisal_rule_code_list_version_to CS,"
        "       S container AT)"
        " OR EXISTS(S container AU, AU is SEDAArchiveUnit)"
    ),
    ("SEDASeqStorageRuleRule", "seda_rule"): (
        "O in_scheme CS, CACLV seda_storage_rule_code_list_version_from AT, "
        "CACLV seda_storage_rule_code_list_version_to CS,"
        "S container AT"
    ),
    # others
    "seda_keyword_reference_to": (
        "O in_scheme CS, S seda_keyword_reference_to_scheme CS"
    ),
}
RTYPE_CARDS = {
    "seda_archive_unit": "?*",
    "seda_binary_data_object": "?*",
    "seda_comment": "1?",
    "seda_custodial_history_item": "1*",
    "seda_description": "1?",
    "seda_description_level_to": "1*",
    "seda_format_id_to": "**",
    "seda_mime_type_to": "**",
    "seda_physical_data_object": "?*",
    "seda_title": "11",
}
RTYPE_CARD = {
    "seda_custodial_history_item": "*",
}
_CARD_TO_CARDS = {
    "1": ["1"],
    "?": ["0..1", "1"],
    "+": ["1..n", "1"],
    "*": ["0..1", "0..n", "1", "1..n"],
}

# Map RQL expression to retrieve matching concept scheme depending on the container type.
# Client code (e.g. in `views/dataobject`) expect scheme to be mapped to the "CS" variable
# and container's eid will be specified using a "container" query argument.
SCHEME_FROM_CONTAINER = {
    "seda_acquisition_information_to": {
        "SEDAArchiveTransfer": (
            "CACLV seda_acquisition_information_code_list_version_from AT, "
            "CACLV seda_acquisition_information_code_list_version_to CS, "
            "AT eid %(container)s"
        ),
    },
    "seda_mime_type_to": {
        "SEDAArchiveTransfer": (
            "CACLV seda_mime_type_code_list_version_from AT, "
            "CACLV seda_mime_type_code_list_version_to CS, "
            "AT eid %(container)s"
        ),
        "SEDAArchiveUnit": 'CS scheme_relation_type RT, RT name "seda_mime_type_to"',
    },
    "seda_encoding_to": {
        "SEDAArchiveTransfer": (
            "CACLV seda_encoding_code_list_version_from AT, "
            "CACLV seda_encoding_code_list_version_to CS, "
            "AT eid %(container)s"
        ),
        "SEDAArchiveUnit": 'CS scheme_relation_type RT, RT name "seda_encoding_to"',
    },
    "seda_format_id_to": {
        "SEDAArchiveTransfer": (
            "CACLV seda_file_format_code_list_version_from AT, "
            "CACLV seda_file_format_code_list_version_to CS, "
            "AT eid %(container)s"
        ),
    },
    "seda_data_object_version_to": {
        "SEDAArchiveTransfer": (
            "CACLV seda_data_object_version_code_list_version_from AT, "
            "CACLV seda_data_object_version_code_list_version_to CS, "
            "AT eid %(container)s"
        ),
    },
    "seda_type_relationship": {
        "SEDAArchiveTransfer": (
            "CACLV seda_relationship_code_list_version_to CS," "AT eid %(container)s"
        ),
    },
    ("SEDABinaryDataObject", "seda_algorithm"): {
        "SEDAArchiveTransfer": (
            "CACLV seda_message_digest_algorithm_code_list_version_from AT, "
            "CACLV seda_message_digest_algorithm_code_list_version_to CS, "
            "AT eid %(container)s"
        ),
        "SEDAArchiveUnit": (
            'CS scheme_relation_type RT, RT name "seda_algorithm", '
            'CS scheme_entity_type ET, ET name "SEDABinaryDataObject"'
        ),
    },
    ("SEDACompressed", "seda_algorithm"): {
        "SEDAArchiveTransfer": (
            "CACLV seda_compression_algorithm_code_list_version_from AT, "
            "CACLV seda_compression_algorithm_code_list_version_to CS,"
            "AT eid %(container)s"
        ),
    },
}


def xsy_mapping(tagname="ArchiveTransfer"):
    mapping = XSYMapping()
    mapping.collect(XSDMMapping(tagname))
    return mapping


def yams_cardinality(minimum, maximum):
    if minimum == 0 and maximum == 1:
        return "?"
    if minimum == 1 and maximum == 1:
        return "1"
    if minimum == 0 and maximum == graph_nodes.INFINITY:
        return "*"
    if minimum == 1 and maximum == graph_nodes.INFINITY:
        return "+"
    assert False


class ETypeMapping:
    """Map an element to its entity type.

    * `element` is the mapped XSD element

    * `etype` is the name of the entity type

    * `card` is a cardinality if the entity should have a `user_cardinality` attribute, else None
    """

    def __init__(self, etype=None, cards=None):
        self.etype = etype
        self.cards = cards
        self.attributes = {}

    def __repr__(self):
        return f"<{self.etype}>"


class RdefMapping:
    """Map a relation to a simple relation definition."""

    def __init__(
        self,
        subjtype,
        rtype,
        objtypes,
        card,
        composite=None,
        inlined=False,
        alias=None,
        desc=(),
        element_name=None,
    ):
        assert len(card) == 2, card
        self.subjtype = subjtype
        self.rtype = rtype
        self.objtypes = objtypes
        self.card = card
        self.composite = composite
        self.inlined = inlined
        self.alias = alias
        self.desc = desc  # list of text collected under xsd:documentation
        self.element_name = element_name  # XML tag name

    @property
    def final(self):
        return next(iter(self.objtypes)) in BASE_TYPES

    def __repr__(self):
        return f"<{self.subjtype} {self.rtype} {self.objtypes}>"


class XSYMapping:
    """XSD 2 Yams mapping: an ordered representation of entity types and relation definitions to be
    defined.
    """

    def __init__(self):
        self.etypes = {}
        self.ordered = []
        self._rdefs = {}

    def collect(self, xsdm_mapping):  # noqa
        """Collect information for future entity types, relation types and relation definitions from
        the given `XSDMMapping` instance.
        """
        for element, etype, child_defs in xsdm_mapping:
            if etype in EXT_ETYPES or etype in BASE_TYPES:
                assert not child_defs, (
                    "unexpected child_defs",
                    element,
                    etype,
                    child_defs,
                )
                continue
            emapping = self.map_etype(etype)
            for occ, path in child_defs:
                if not path:
                    continue  # jumped element
                if path[0][2] in SKIP_ETYPES:
                    continue
                current_emapping = emapping
                is_complex = len(path) == 2
                end_etype = path[-1][2]
                # cardinality specified in the XSD
                card = yams_cardinality(occ.minimum, occ.maximum)
                # cardinality fixed according to parent (set to 0 if element is in a choice)
                fixedcard = yams_cardinality(occ.fixed_minimum, occ.maximum)
                for i, (rtype, role, target_etype, rdef_options) in enumerate(path):
                    if target_etype in BASE_TYPES:
                        assert role == "subject"
                        if rtype not in SKIP_ATTRS:
                            current_emapping.attributes[rtype] = target_etype
                        continue
                    # we want seda_uri for SEDAAltBinaryDataObjectAttachment but not for Attachment
                    # (it should be systematically added in XSD for the later)
                    if rtype == "seda_uri" and emapping.etype == "SEDAAttachment":
                        continue
                    rdef_options["desc"] = getattr(occ.target, "desc", ())
                    rdef_options["element_name"] = getattr(
                        occ.target, "local_name", None
                    )
                    if role == "subject":
                        subjtype = current_emapping.etype
                        objtype = target_etype
                        rdef_options["alias"] = _alias(subjtype, rtype)
                    else:
                        subjtype = target_etype
                        objtype = current_emapping.etype
                        rdef_options["alias"] = _alias(objtype, rtype)
                    # compute rdef cardinalities depending on composite and fixedcard
                    composite = rdef_options.get("composite")
                    if rtype in RTYPE_CARDS:
                        rdef_options["card"] = RTYPE_CARDS[rtype]
                        if RTYPE_CARDS[rtype][0] not in "?1":
                            rdef_options["inlined"] = False
                    elif composite == "subject":
                        rdef_options["card"] = fixedcard + "1"
                    elif composite == "object":
                        rdef_options["card"] = "1" + fixedcard
                    elif fixedcard in "1?":
                        rdef_options["card"] = "?*"
                    else:
                        rdef_options["card"] = "**"
                    if rtype in RTYPE_CARD:
                        card = RTYPE_CARD[rtype]
                    # create mapping for intermediary entity type if necessary
                    if target_etype not in EXT_ETYPES and not isinstance(
                        target_etype, tuple
                    ):
                        cards = _CARD_TO_CARDS[card]
                        current_emapping = self.map_etype(target_etype, cards)
                    # merge rdef with a previously added one (only for complex relations) or create
                    # a new one
                    if is_complex and i == 1 and (end_etype, rtype) in self._rdefs:
                        ref_mapping = self._rdefs[(end_etype, rtype)]
                        _merge_mapping(
                            ref_mapping, subjtype, rtype, objtype, composite, fixedcard
                        )
                    elif is_complex and i == 0 and (rtype, end_etype) in self._rdefs:
                        ref_mapping = self._rdefs[(rtype, end_etype)]
                        _merge_mapping(
                            ref_mapping, subjtype, rtype, objtype, composite, fixedcard
                        )
                    # occ is element.parent when occurence is reused for element's textual content
                    elif occ is not element.parent:
                        ref_mapping = self.map_rdef(
                            subjtype, rtype, objtype, **rdef_options
                        )
                        if is_complex:
                            key = (end_etype, rtype) if i == 1 else (rtype, end_etype)
                            self._rdefs[key] = ref_mapping

    def map_etype(self, etype, cards=None):
        """Register an entity type."""
        emapping = ETypeMapping(etype, cards)
        if etype in self.etypes:
            emapping = self.etypes[etype]
            if cards is not None:
                if emapping.cards is None:
                    emapping.cards = cards
                else:
                    cards = set(cards)
                    emapping_cards = set(emapping.cards)
                    if cards - emapping_cards:
                        print(
                            "# XXX extending cards for",
                            emapping,
                            cards - emapping_cards,
                        )
                        emapping.cards = sorted(emapping_cards | cards)
        else:
            self.etypes[emapping.etype] = emapping
            self.ordered.append(emapping)
        return emapping

    def map_rdef(
        self,
        subjtype,
        rtype,
        objtypes,
        card,
        composite=None,
        inlined=False,
        alias=None,
        desc=(),
        element_name=None,
    ):
        """Register a relation definition."""
        objtypes = _ensure_set(objtypes)
        mapping = RdefMapping(
            subjtype,
            rtype,
            objtypes,
            card=card,
            composite=composite,
            inlined=inlined,
            alias=alias,
            desc=desc,
            element_name=element_name,
        )
        self.ordered.append(mapping)
        return mapping


def _ensure_set(value):
    """Given a string or a tuple, return a set."""
    if isinstance(value, set):
        return value
    elif isinstance(value, tuple):
        return set(value)
    else:
        return {value}


def _merge_mapping(ref_mapping, subjtype, rtype, objtype, composite, card=None):
    assert composite == ref_mapping.composite, (ref_mapping, composite)
    if card is not None:
        if not card == ref_mapping.card[1]:
            print(
                "# XXX unsupported merge because of incompatible cardinality",
                subjtype,
                rtype,
                objtype,
                card,
                ref_mapping,
                ref_mapping.card,
            )
        ref_mapping.objtypes.update(_ensure_set(objtype))
    else:
        assert ref_mapping.objtypes == _ensure_set(objtype)


# code generation ##################################################################################


class CodeGenerator:
    """Abstract class for code generator"""

    @classmethod
    def main(cls):
        """A main implementation to generate code"""
        import sys

        if len(sys.argv) < 2:
            mapping = xsy_mapping()
        else:
            assert len(sys.argv) == 2
            mapping = xsy_mapping(sys.argv[1])
        cls().generate(mapping, sys.stdout)

    def generate(self, mapping, stream, with_header=True):
        """Generator entry point: write generated code for :class:`XSYMapping`
        into the given stream"""
        if with_header:
            stream.write(_PY_HEADER)
        self._generate(mapping, stream)

    def _generate(self, mapping, stream):
        """Override me in concret class"""
        raise NotImplementedError()

    def _callback(self, prefix, mapping_element):
        """Call method named according to `prefix` and `mapping_element`'s class and return its
        result or an empty tuple if the method doesn't exist
        """
        callback = f"{prefix}_{un_camel_case(mapping_element.__class__.__name__)}"
        try:
            method = getattr(self, callback)
        except AttributeError:
            return ()
        return method(mapping_element)


class YamsSchemaGenerator(CodeGenerator):
    """Yams schema generator"""

    def _generate(self, mapping, stream):
        stream.write(
            """from yams.buildobjs import EntityType, RelationDefinition
from yams.buildobjs import String, Boolean, Date
from cubicweb.schema import RQLConstraint
from cubicweb_seda.schema import seda_profile_element


"""
        )
        self._processed_complex_rdef = set()
        for mapping_element in mapping.ordered:
            mapping_code = self._callback("code_for", mapping_element)
            stream.write(mapping_code)

    def code_for_e_type_mapping(self, mapping):
        if mapping.cards:
            cards = mapping.cards
            annotable = len(cards) > 1
            default_card = "1"
            code = f"""\
@seda_profile_element(cardinalities={cards}, default_cardinality='{default_card}',
                      annotable={annotable})
"""
        else:
            # SEDAArchiveTransfert
            code = "@seda_profile_element()\n"
        code += f'''class {mapping.etype}(EntityType):
    u"""""\"
'''
        for attrname, attrtype in sorted(mapping.attributes.items()):
            if attrname == "id":
                continue
            args = ""
            if attrtype == "String":
                args = "fulltextindexed=True"
            code += f"    {attrname} = {attrtype}({args})\n"
        return code + "\n\n"

    def code_for_rdef_mapping(self, mapping):
        rtype = mapping.rtype
        alias = mapping.alias
        if not alias:
            alias = _alias(mapping.subjtype, rtype)
        constraint = _constraint(mapping.subjtype, mapping.rtype)
        if len(mapping.objtypes) == 1:
            objtypes = repr(next(iter(mapping.objtypes)))
        else:
            objtypes = tuple(sorted(mapping.objtypes))
        composite = repr(mapping.composite)
        return f"""class {alias}(RelationDefinition):
    name = '{mapping.rtype}'
    subject = '{mapping.subjtype}'
    object = {objtypes}
    cardinality = '{mapping.card}'
    composite = fulltext_container = {composite}
    inlined = {mapping.inlined}
    constraints = [{constraint}]

"""


def _constraint(etype, rtype):
    try:
        expr = RDEF_CONSTRAINTS[(etype, rtype)]
    except KeyError:
        try:
            expr = RDEF_CONSTRAINTS[rtype]
        except KeyError:
            return ""
    return f"RQLConstraint({expr.format(subjtype=etype)!r})"


def _alias(etype, rtype):
    if isinstance(etype, tuple):
        etype = "".join(sorted(etype))
    return (un_camel_case(etype) + "_" + rtype).replace("seda_", "")


_PY_HEADER = '''# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""THIS FILE IS GENERATED FROM SEDA 2.2 XSD FILES, DO NOT EDIT"""

'''


if __name__ == "__main__":
    YamsSchemaGenerator.main()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda views for ArchiveTransfer"""

from logilab.common.decorators import monkeypatch
from logilab.common.registry import objectify_predicate

from cubicweb import _, tags
from cubicweb.predicates import is_instance
from cubicweb_web import formwidgets as fw
from cubicweb_web.views import tabs, uicfg, reledit

from cubicweb_relationwidget import views as rwdg

from ..xsd2yams import RULE_TYPES
from ..entities import full_seda2_profile, simplified_profile, parent_and_container
from . import rtags_from_xsd_element, rtags_from_rtype_role_targets
from . import clone, viewlib
from . import uicfg as sedauicfg  # noqa - ensure those rules are defined first

at_ordered_fields = [
    ("seda_archival_agency", "subject"),
    ("seda_transferring_agency", "subject"),
    ("seda_transfer_request_reply_identifier", "object"),
    ("seda_comment", "object"),
    ("seda_signature", "object"),
    ("seda_archival_agreement", "object"),
]

pvs = uicfg.primaryview_section
pvdc = uicfg.primaryview_display_ctrl
rec = uicfg.reledit_ctrl
afs = uicfg.autoform_section
affk = uicfg.autoform_field_kwargs

pvs.tag_attribute(("SEDAArchiveTransfer", "title"), "hidden")
pvs.tag_attribute(("SEDAArchiveTransfer", "simplified_profile"), "hidden")
pvs.tag_attribute(("SEDAArchiveTransfer", "compat_list"), "hidden")

for rtype, role in at_ordered_fields:
    if rtype.endswith("agency"):
        assert role == "subject"
        pvs.tag_subject_of(("SEDAArchiveTransfer", rtype, "*"), "attributes")
        rec.tag_subject_of(
            ("SEDAArchiveTransfer", rtype, "*"),
            {
                "rvid": "seda.reledit.simplelink",
                "novalue_label": _("<no value specified>"),
            },
        )
    else:
        assert role == "object"
        pvs.tag_object_of(("*", rtype, "SEDAArchiveTransfer"), "attributes")
        if rtype in ("seda_archival_agreement", "seda_comment"):
            simplified_section = "attributes"
        else:
            simplified_section = "hidden"
        rec.tag_object_of(
            ("*", rtype, "SEDAArchiveTransfer"),
            {"rvid": "seda.reledit.text", "novalue_label": " "},
        )
        # hide extra-fields from main form, as their appearance depends on the simplified_profile
        # switch
        afs.tag_object_of(("*", rtype, "SEDAArchiveTransfer"), "main", "hidden")

affk.set_field_kwargs("SEDAArchiveTransfer", "title", widget=fw.TextInput({"size": 80}))
affk.set_fields_order(
    "SEDAArchiveTransfer",
    ["title", "simplified_profile", "user_annotation"] + at_ordered_fields,
)
pvdc.set_fields_order(
    "SEDAArchiveTransfer", ["title", "user_annotation"] + at_ordered_fields
)


# don't show inheritance control on top level rules
@objectify_predicate
def top_level_rule(cls, req, rset=None, entity=None, **kwargs):
    """Predicate returning 1 score if rule is linked to the transfer, not to an archive unit."""
    if entity is None:
        entity = rset.one()
    parent = parent_and_container(entity)[0]
    if parent.cw_etype == "SEDAArchiveTransfer":
        return 1
    return 0


top_level_rule_afs = afs.derive(
    __name__,
    is_instance(*[f"SEDA{rule_type.capitalize()}Rule" for rule_type in RULE_TYPES])
    & top_level_rule(),
)
for rule_type in RULE_TYPES:
    etype = f"SEDA{rule_type.capitalize()}Rule"
    rtype = f"seda_alt_{rule_type}_rule_prevent_inheritance"
    top_level_rule_afs.tag_subject_of((etype, rtype, "*"), "main", "hidden")


class ArchiveTransferTabbedPrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance("SEDAArchiveTransfer")
    tabs = [
        "main_tab",
        _("seda_at_code_list_versions_tab"),
        _("seda_management_tab"),
        _("seda_data_objects_tab"),
        _("seda_archive_units_tab"),
        _("seda_at_related_transfers_tab"),
        _("seda_at_diagnose_tab"),
    ]

    def entity_call(self, entity, **kwargs):
        super().entity_call(entity, **kwargs)
        rwdg.bootstrap_dialog(self.w, self._cw._, clone._import_div_id(entity))


class ArchiveTransferCodeListVersionsTab(viewlib.PrimaryTabWithoutBoxes):
    """Tab for code list versions information of an archive transfer."""

    __regid__ = "seda_at_code_list_versions_tab"
    __select__ = is_instance("SEDAArchiveTransfer") & full_seda2_profile()

    rsection, display_ctrl = rtags_from_xsd_element(
        "SEDAArchiveTransfer", "CodeListVersions"
    )


class ArchiveTransferManagementTab(viewlib.PrimaryTabWithoutBoxes):
    """Main tab for management data of an archive transfer"""

    __regid__ = "seda_management_tab"
    __select__ = is_instance("SEDAArchiveTransfer")

    rtype_role_targets = [
        ("seda_storage_rule", "object", None),
        ("seda_appraisal_rule", "object", None),
        ("seda_access_rule", "object", None),
        ("seda_dissemination_rule", "object", None),
        ("seda_reuse_rule", "object", None),
        ("seda_hold_rule", "object", None),
        ("seda_classification_rule", "object", None),
        ("seda_need_authorization", "object", None),
        ("seda_service_level", "object", None),
        ("seda_archival_profile", "object", None),
        ("seda_acquisition_information_from", "object", None),
        ("seda_legal_status_from", "object", None),
        ("seda_originating_agency_identifier", "object", None),
        ("seda_submission_agency_identifier", "object", None),
    ]
    rsection, display_ctrl = rtags_from_rtype_role_targets(
        "SEDAArchiveTransfer", rtype_role_targets
    )


class SimplifiedArchiveTransferManagementTab(ArchiveTransferManagementTab):
    __select__ = ArchiveTransferManagementTab.__select__ & simplified_profile()

    rtype_role_targets = [
        ("seda_appraisal_rule", "object", None),
        ("seda_access_rule", "object", None),
    ]
    rsection, display_ctrl = rtags_from_rtype_role_targets(
        "SEDAArchiveTransfer", rtype_role_targets
    )


class ArchiveTransferDataObjectsTab(viewlib.SubObjectsTab):
    """Tab for data objects of an archive transfer"""

    __regid__ = "seda_data_objects_tab"
    __select__ = is_instance("SEDAArchiveTransfer") & full_seda2_profile()

    rtype_role_targets = [
        ("seda_binary_data_object", "object", "SEDABinaryDataObject"),
        ("seda_physical_data_object", "object", "SEDAPhysicalDataObject"),
    ]

    _(
        "creating SEDABinaryDataObject (SEDABinaryDataObject seda_binary_data_object "
        "SEDAArchiveTransfer %(linkto)s)"
    )
    _(
        "creating SEDAPhysicalDataObject (SEDAPhysicalDataObject seda_physical_data_object "
        "SEDAArchiveTransfer %(linkto)s)"
    )


class ArchiveTransferArchiveUnitsTab(viewlib.SubObjectsTab):
    """Tab for archive units of an archive transfer"""

    __regid__ = "seda_archive_units_tab"
    __select__ = is_instance("SEDAArchiveTransfer")

    rtype_role_targets = [("seda_archive_unit", "object", None)]

    _(
        "creating SEDAArchiveUnit (SEDAArchiveUnit seda_archive_unit "
        "SEDAArchiveTransfer %(linkto)s)"
    )

    def display_add_button(self, entity):
        rtype_roles = [(rtype, role) for rtype, role, _ in self.rtype_role_targets]
        links = viewlib.add_links_from_rtype_roles(
            entity, rtype_roles, self.url_params(entity)
        )
        if links:
            import_url = clone.import_unit_url(self._cw, entity)
            links.append(tags.a(self._cw._("import_unit"), href=import_url))
        viewlib.display_add_button(self.w, links, self._cw.__("add"))

    def url_params(self, entity):
        return {
            "__redirectparams": "tab=" + self.__regid__,
            "unit_type": "unit_content",
        }


class ArchiveTransferRelatedTransfersTab(viewlib.SubObjectsTab):
    """Tab for previous transfers related to an archive transfer"""

    __regid__ = "seda_at_related_transfers_tab"
    __select__ = is_instance("SEDAArchiveTransfer") & full_seda2_profile()

    rtype_role_targets = [("seda_related_transfer_reference", "object", None)]

    _(
        "creating SEDARelatedTransferReference (SEDARelatedTransferReference "
        "seda_related_transfer_reference SEDAArchiveTransfer %(linkto)s)"
    )


class ArchiveTransferDiagnoseTab(viewlib.SubObjectsTab):
    """Tab to diagnose supported format of an archive transfer"""

    __regid__ = "seda_at_diagnose_tab"
    __select__ = is_instance("SEDAArchiveTransfer")

    def entity_call(self, entity):
        self.w('<p class="bg-info">')
        formats = entity.formats_compat
        if "simplified" in formats:
            formats.remove("simplified")
        self.w(self._cw._("Supported formats: {0}.").format(", ".join(sorted(formats))))
        self.w("</p>")
        doctor = entity.cw_adapt_to("ISEDACompatAnalyzer")
        data = [
            (
                tags.a(e.entity.dc_title(), href=e.entity.absolute_url(tab=e.tab_id)),
                ", ".join(e.impacted_formats),
                self._cw._(e.message),
            )
            for e in doctor.detect_problems()
        ]
        if data:
            self.wview(
                "pyvaltable",
                pyvalue=data,
                headers=(
                    self._cw._("Entity"),
                    self._cw._("Impacted formats"),
                    self._cw._("Message"),
                ),
            )
        else:
            self.w(self._cw._("No problem detected, congratulation!"))


@monkeypatch(reledit.AutoClickAndEditFormView)
def _compute_formid_value(self, rschema, role, rvid, formid):
    """Overriden to give rtype/role information to the view"""
    related_rset = self.entity.related(rschema.type, role)
    if related_rset:
        value = self._cw.view(
            rvid, related_rset, initargs={"rtype": rschema.type, "role": role}
        )  # path is here
    else:
        value = self._compute_default_value(rschema, role)
    if not self._should_edit_relation(rschema, role):
        return None, value
    return formid, value
